﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;
using UiO.TRAMP.RendezvousProtocol;
using UiO.TRAMP.MSRP;
using UiO.TRAMP.Logger;
using System.Net;
using UiO.TRAMP.Transport;
using UiO.TRAMP.State;
using System.Threading;

namespace UiO.TRAMP.DataTransfer
{
    class DataTransferClass
    {
        private LoggerClass Log;
        private RendezvousProtocolClass rendezvousProtcol;

        private RandezvousProtocolParsingResult result;
        private SIPPayloadState state;

        private String selectedLocalIPAddress;
        private Boolean source;

        private MSRPClass msrp;

        private ManualResetEvent syncNegotiationAndDataTransfer;

        private Stopwatch rendezvousSw = new Stopwatch();
        private Stopwatch dataTransferSw = new Stopwatch();

        private TimeSpan rendezvousTimeSpan;
        private TimeSpan dataTransferTimeSpan;

        private byte[] payload;
        private int payloadLength;

        public DataTransferClass(LoggerClass Log, String selectedLocalIPAddress, Object gui) {
            this.Log = Log;
            this.msrp = new MSRPClass(Log, gui);
            this.selectedLocalIPAddress = selectedLocalIPAddress;            
            syncNegotiationAndDataTransfer = new ManualResetEvent(false);
        }

        private void Abort() {
            ReleaseResources();
            Thread.CurrentThread.Abort();
        }

        /// <summary>
        /// We will be using the existing connection (TCPConnection class) from negotiation protocol 
        /// in order to save the TCP connection establishment time
        /// </summary>
        /// <param name="source"></param>
        public void TransferData(Boolean source, TCPConnection tcpConn, IPAddress ipAddress) {
            this.source = source;
            this.rendezvousProtcol = new RendezvousProtocolClass(Log, tcpConn);
            result = RendezvousProtocol.RandezvousProtocolParsingResult.OK; // we do not add any fancy decision algorithm here, but we might consider to return NOT FOUND in some cases
            if (source){
                CreateLocalPartOfDataTransferSession(selectedLocalIPAddress);
                InitializeRandezvousPayload(selectedLocalIPAddress);
                //Rendezvous protocol start
                rendezvousSw.Start();
                SendRendezvousRequest(ipAddress, selectedLocalIPAddress);
                ReceiveResposeAfterRendezvousRequest();
                CreateRemotePartOfDataTransferSession();
                AckResponseToRendezvousRequest();
                rendezvousSw.Stop();
                //Rendezvous protocol end
                dataTransferSw.Start();
                SendData();
                dataTransferSw.Stop();

                // Write performance results to the log
                if (Log.WriteToLog != true) {
                    Log.WriteToLog = true;
                    Log.WriteLog("UiO.TRAMP.DataTransfer.TransferData(): Writing the rendezvous protocol duration to the log file.");
                    rendezvousTimeSpan = rendezvousSw.Elapsed;
                    Log.WriteLog("UiO.TRAMP.DataTransfer.TransferData(): The rendezvous protocol took '" + rendezvousTimeSpan.TotalMilliseconds + "' ms.");
                    Log.WriteLog("UiO.TRAMP.DataTransfer.TransferData(): Writing data transfer duration to the log file.");
                    dataTransferTimeSpan = dataTransferSw.Elapsed;
                    Log.WriteLog("UiO.TRAMP.DataTransfer.TransferData(): The data transfer took '" + dataTransferTimeSpan.TotalMilliseconds + "' ms.");
                    Log.WriteToLog = false;
                }

                SendRendezvousSessionCancellationRequest();
                ReceiveResponseAfterRendezvousSessionCancellationRequest();
            } else {
                CreateLocalPartOfDataTransferSession(selectedLocalIPAddress);
                BindAndListenMSRP();
                ReceiveRendezvousRequest();
                CreateRemotePartOfDataTransferSession();                
                InitializeRandezvousPayload(selectedLocalIPAddress);
                SendResponseToRendezvousRequest();
                ReceiveAckOnResponseToRendezvousRequest();                
                /*
                    There is an apsolute need for synchronization here!
                 */
                // WAIT FOR BIND                
                if (!msrp.aborted) {
                    Log.WriteDebug("UiO.TRAMP.DataTransfer.TransferData(): Waiting for the connection request from the remote peer in order to continue data transfer receive.");
                    syncNegotiationAndDataTransfer.WaitOne();
                } else {
                    Log.WriteLogWarning("UiO.TRAMP.DataTransfer.TransferData(): The connection was aborted on the receiving thread. Unable to continue. Aborting the main thread.");
                    Abort();
                }
                Log.WriteDebug("UiO.TRAMP.DataTransfer.TransferData(): Connection to the remote host is established. Continuing to the data transfer.");
                ReceiveData();
                ReceiveRendezvousSessionCancelationRequest();
                SendResponseToRendezvousSessionCancelationRequest();
            }
        }

        private void SendRendezvousSessionCancellationRequest() {
            // Sends SIP BYE and releases all dialogs
            rendezvousProtcol.SendSessionCancellationRequest();            
        }

        private void ReceiveResponseAfterRendezvousSessionCancellationRequest() {
            rendezvousProtcol.ReceiveResponseAfterSessionCancellationRequest();
        }

        private void PassiveSessionRelease() { 
            // waits for SIP BYE from remote peer and releases all dialogs upon its arrival
            // maybe timers would be appropriate here?
        }

        private void BindAndListenMSRP() {
            Thread listeningThread = new Thread(new ParameterizedThreadStart(Bind)); ;
            listeningThread.Start(this);
        }

        private void SendData() {
            msrp.TransferData();
        }

        public void Bind(Object state) {
            DataTransferClass dt = (DataTransferClass)state;

            dt.msrp.BindAndListenMSRPSync();

            dt.Log.WriteDebug("UiO.TRAMP.DataTransfer.Bind(): Signalling the main thread so that the data transfer can begin.");
            syncNegotiationAndDataTransfer.Set();
                                   
        }

        public void ReleaseResources() {
            Log.WriteLog("UiO.TRAMP.DataTransfer.ReleaseResources(): Releasing network resources in the 'Data Transfer' layer.");
            rendezvousProtcol.ReleaseResources();
            msrp.ReleaseResources();
        }

        private void ReceiveRendezvousSessionCancelationRequest()
        {
            rendezvousProtcol.ReceiveSessionCancelationRequest(); 
        }

        private void SendResponseToRendezvousSessionCancelationRequest() {
            rendezvousProtcol.SendResponseToSessionCancelationRequest();
        }

        private void ReceiveData() {            
            msrp.ReceiveTransferedData();            
        }

        private void CreateLocalPartOfDataTransferSession (String selectedLocalIPAddress){
            msrp.CalculateLocalMSRPUri(selectedLocalIPAddress);
        }

        private void CreateRemotePartOfDataTransferSession() {
            /*
                As far as we are concerned right now, we do not need anything from the received SDP other than msrpURI of the remote end, since we do not have any 
             * decision making algorithm here that could accept or deny rendezvous request. Therefore a very simple parser is defined here to extract a remote host's msrpUri            
             */
            Boolean foundMSRPpathAttribute = false;
            String remoteUri = "";

            String rcvdSDP = new UTF8Encoding().GetString(state.content);
            String[] sdpLines = rcvdSDP.Split(new string[] { Environment.NewLine }, StringSplitOptions.None);

            Log.WriteLog("UiO.TRAMP.DataTransfer.CreateRemotePartOfDataTransferSession(): Parsing received SDP document.");

            foreach(String sdpLine in sdpLines) {
                if (sdpLine.StartsWith("a=path")){
                    foundMSRPpathAttribute = true;
                    remoteUri = sdpLine.Substring(7, sdpLine.Length-7); // not exactly perfect but it works ...
                    Log.WriteLog("UiO.TRAMP.DataTransfer.CreateRemotePartOfDataTransferSession(): Found remote msrpURI in the received SDP document '" + remoteUri + "'.");
                    break;
                }
            }

            if (!foundMSRPpathAttribute) {
                Log.WriteLogWarning("UiO.TRAMP.DataTransfer.CreateRemotePartOfDataTransferSession(): Unable to find path attribute in the received SDP document. Aborting.");
                if (!source) {
                    result = RendezvousProtocol.RandezvousProtocolParsingResult.NOTFOUND;
                } else {
                    Abort();
                }
            }

            msrp.UpdateRemoteMSRPUri(remoteUri);
        }

        private void InitializeRandezvousPayload(String selectedLocalIPAddress) {
            /*
                Since the SDP use in this context is very simple and limited we will not be making a separate SDP class 
             * to accomodate SDP. If this at some later point turns out to be a neccessety, however, this class will be created
             */
            StringBuilder sb = new StringBuilder();

            sb.Append("v=0");
            sb.Append("\r\n");
            sb.Append("o=- " + msrp.GetRandomNumberString(10) + " " + msrp.GetRandomNumberString(10) + " " + "IN IP4 " + selectedLocalIPAddress);
            sb.Append("\r\n");
            sb.Append("s=-");
            sb.Append("\r\n");
            sb.Append("c=IN IP4 " + selectedLocalIPAddress);
            sb.Append("\r\n");
            sb.Append("t=0 0");
            sb.Append("\r\n");
            sb.Append("m=message " + MSRPClass.c_default_msrp_port + " TCP/MSRP *");
            sb.Append("\r\n");
            sb.Append("a=accept-types:application/octet-stream");
            sb.Append("\r\n");
            sb.Append("a=accept-wrapped-types:*");
            sb.Append("\r\n");
            sb.Append("a=path:" + msrp.localMSRPUri.ToString());
            sb.Append("\r\n");

            Log.WriteDebug("UiO.TRAMP.DataTransfer.InitializeRandezvousPayload(): Constructed rendezvous payload is: \r\n\r\n" + sb.ToString());

            payload = (new UTF8Encoding()).GetBytes(sb.ToString());
            payloadLength = payload.Length;            
        }

        private void SendRendezvousRequest(IPAddress ipAddress, String selectedLocalIPAddress) {            

            Log.WriteLog("UiO.TRAMP.DataTransfer.SendRendezvousRequest(): Sending rendezvous request to the remote host.");
            rendezvousProtcol.SendRendezvousRequest(ipAddress, selectedLocalIPAddress, payload, payloadLength);
            Log.WriteDebug("UiO.TRAMP.Negotiation.SendNegotiationRequest(): Returned from SendRendezvousRequest function.");
        }

        private void ReceiveResposeAfterRendezvousRequest() {
            Log.WriteLog("UiO.TRAMP.DataTransfer.ReceiveResposeAfterRendezvousRequest(): Receiving response after rendezvous request.");          

            rendezvousProtcol.ReceiveResposeAfterRendezvousRequest();
            Log.WriteLogSuccess("UiO.TRAMP.DataTransfer.ReceiveResposeAfterRendezvousRequest(): Received response to rendezvous request.");

            // populate state variable in order to receive payload and other state data
            state = rendezvousProtcol.payloadState;

            /*
             *  state variable, the one that carries the response payload to the "negotiation" layer will also have the information about the type of response 
             *  that has been received. This will be used to populate negotiationAlgorithm variable which in turn will equip  AckResponseToNegotiationRequest() to either continue to 
             *  data transfer (200 OK response) or abort (all other responses)
             */
        }

        private void AckResponseToRendezvousRequest() {
            Log.WriteLog("UiO.TRAMP.DataTransfer.AckResponseToRendezvousRequest(): Acknowledging response to negotiation request.");
            rendezvousProtcol.SendAckToRendezvousResponse();

            switch (state.rType) 
            {
                case SIPPayloadState.receivedResponseType.OK:
                    Log.WriteLogSuccess("UiO.TRAMP.DataTransfer.AckResponseToRendezvousRequest(): Renedzvous is finished. Moving on to the data transfer.");
                    break;
                case SIPPayloadState.receivedResponseType.NOTFOUND:
                    Log.WriteLogWarning("UiO.TRAMP.DataTransfer.AckResponseToRendezvousRequest(): Acknowledged negative rendezvous response due to non-existant resources. Aborting.");
                    // Stop the negotiation process
                    Abort();
                    break;                
                default:
                    break;
            }
        }

        private void ReceiveRendezvousRequest() {
            Log.WriteLog("UiO.TRAMP.DataTransfer.ReceiveRendezvousRequest(): Entering negotiation protocol layer to listen for negotiation request.");            

            rendezvousProtcol.ReceiveRendezvousRequest();            
            // extract state
            state = rendezvousProtcol.payloadState;

            Log.WriteDebug("UiO.TRAMP.DataTransfer.ReceiveRendezvousRequest(): Negotiation layer returned negotiation request. Running negotiation algorithm.");
        }

        private void SendResponseToRendezvousRequest() {
            switch (result)
            {
                case RandezvousProtocolParsingResult.OK:
                    Log.WriteLogSuccess("UiO.TRAMP.DataTransfer.SendResponseToRendezvousRequest(): Data transfer is possible. Returning positive response.");
                    rendezvousProtcol.SendPositiveRendezvousResponse(payload, payloadLength);
                    Log.WriteDebug("UiO.TRAMP.DataTransfer.SendResponseToRendezvousRequest(): Positive response sent.");
                    break;
                case RandezvousProtocolParsingResult.NOTFOUND:
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.SendResponseToRendezvousRequest(): Data transfer is not possible. Returning negative response.");
                    rendezvousProtcol.SendNegativeRendezvousResponse(payload, payloadLength);
                    Log.WriteDebug("UiO.TRAMP.DataTransfer.RunNegotiationAlgorithm(): Negative response sent.");
                    break;
                default:
                    Log.WriteLogWarning("UiO.TRAMP.DataTransfer.SendResponseToRendezvousRequest(): Unknown result from rendezvous payload parsing. Aborting.");
                    Abort();
                    break;
            }
        }

        public void ReceiveAckOnResponseToRendezvousRequest()
        {
            rendezvousProtcol.ReceiveAckOnResponseToRendezvousRequest();

            switch (result)
            {
                case RandezvousProtocolParsingResult.OK:
                    Log.WriteLogSuccess("UiO.TRAMP.DataTransfer.ReceiveAckOnResponseToRendezvousRequest(): Received ack on positive response to rendezvous request.");
                    // continuing with data transfer
                    break;
                case RandezvousProtocolParsingResult.NOTFOUND:
                    Log.WriteLogWarning("UiO.TRAMP.DataTransfer.ReceiveAckOnResponseToRendezvousRequest(): Received ack on negative response. Stoping data transfer process.");
                    // Stop the negotiation process
                    Abort();
                    break;                
                default:
                    Log.WriteLogWarning("UiO.TRAMP.DataTransfer.ReceiveAckOnResponseToRendezvousRequest(): Unknown result from rendezvous payload parsing. Aborting.");
                    Abort();
                    break;
            }
        }
    }
}
