﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;

namespace UiO.TRAMP.Logger
{
    /// <summary>
    /// This class represents a logger for the entire project 
    /// We are using lock statement to coordinate access to logger by multiple threads
    /// The most important events are also logged into GUI so that the user can get an overview
    /// There is a difference between debugg mode and normal mode (debug mode gives more information)
    /// </summary>
    public class LoggerClass
    {
        private const string c_datetime_format = "yyyy.MM.dd HH:mm:ss";

        private string LogFilePath;
        private string LogFileFolderName;
        private string LogFileName;

        /// <summary>
        /// We are adding locks to our code since it is certain that multiple
        /// threads will try to write to logg at the same time
        /// </summary>

        private object logWriteLock = new object();

        private Boolean WDebug;
        private Boolean WGUI;
        private Boolean WLog; 

        private System.Windows.Forms.RichTextBox textBox;

        delegate void AsyncGUIUpdate(String sLogLine, GUIEventType eType);

        public LoggerClass (string LogFileName, String LogFileFolderName, System.Windows.Forms.RichTextBox textBox)
        {
            this.LogFileName = LogFileName;
            this.LogFileFolderName = LogFileFolderName;
            this.textBox = textBox;

            this.WDebug = false;
            this.WGUI = true;
            this.WLog = true;

            if (LogFileFolderName.Substring(LogFileFolderName.Length - 1, 1) == @"\") {
                LogFilePath = LogFileFolderName + LogFileName;
            } else {
                LogFilePath = LogFileFolderName + @"\" + LogFileName;
            }
        }

        public Boolean WriteToDebug {

            get {
                return WDebug;
            }
            
            set {
                WDebug = value;
            }
        }

        public Boolean WriteToLog {
            get {
                return WLog;
            }

            set {
                WLog = value;
            }
        }
        
        /// <summary>
        /// Decides whether the events are written to a GUI as well
        /// Both writelog and writeDebug functions can write to GUI if this field is true
        /// </summary>
        public Boolean WriteToGUI {
            get {
                return WGUI;
            }

            set {
                WGUI = value;
            }
        }

        public String LoggFilePath {

            get {
                return LogFilePath;
            }

            set {
                LogFilePath = value;
            }
        }

        public void OpenLogFile () {                        
            try {
                if (!File.Exists(LogFilePath)) {
                    ForceFolder(LogFileFolderName);
                    
                    // create file because it doesn't exist from before
                    using (StreamWriter sw = File.CreateText(LogFilePath)) {
                        sw.WriteLine(DateTime.Now.ToString(c_datetime_format) + " " + "************************* Opening Log *****************************");
                    }
                } else {
                    // append
                    using (StreamWriter sw = new StreamWriter (LogFilePath, true)){
                        sw.WriteLine(DateTime.Now.ToString(c_datetime_format) + " " + "**************** Appending to existing Log File *******************");
                    }
                }

                WriteLog("OpenLogFile(): Successfully opened log file '" + LogFilePath + "'");

            } catch (Exception exp) {
                // nothing to do with this here ...
                System.Windows.Forms.MessageBox.Show(exp.ToString());                        
            }
        }

        private void ForceFolder(String sFolderName) {

            string sPath = "";
            string[] aKeys = sFolderName.Split('\\');  

            foreach (String sKey in aKeys){
                if (sPath.Length > 0) {
                    sPath = sPath + @"\";
                }

                if (sKey.Length == 0) { continue; }

                sPath = sPath + sKey;

                switch (sKey.Substring(sKey.Length -1, 1)) {
                    case ":":
                        break;
                    case ".":
                        break;
                    default:                     
                        if (!Directory.Exists(sPath)) {
                            Directory.CreateDirectory(sPath);                                
                        }                        
                        break;                
                }
            }            
        }

        public void WriteLog(String sLogLine, GUIEventType eType)
        {
            lock (logWriteLock) {
                if(WLog)
                    using (StreamWriter sw = new StreamWriter(LogFilePath, true)) {
                        sw.WriteLine(DateTime.Now.ToString(c_datetime_format) + " [LOG] [TID '" + System.Threading.Thread.CurrentThread.ManagedThreadId + "']: " + sLogLine);
                        if (WGUI)
                            WriteGUI(DateTime.Now.ToString(c_datetime_format) + " [LOG] [TID '" + System.Threading.Thread.CurrentThread.ManagedThreadId + "']: " + sLogLine + "\r\n", eType);
                    }
            }
        }

        public void WriteLog(String sLogLine) {
            WriteLog(sLogLine, GUIEventType.COMMON);
        }

        public void WriteLogWarning (String sLogLine)
        {
            WriteLog(sLogLine, GUIEventType.WARNING);
        }

        public void WriteLogInformational (String sLogLine)
        {
            WriteLog(sLogLine, GUIEventType.INFORMATIONAL);
        }

        public void WriteLogSuccess(String sLogLine)
        {
            WriteLog(sLogLine, GUIEventType.SUCCESS);
        }

        public void WriteDebug(String sLogLine, GUIEventType eType)
        {
            lock (logWriteLock)
            {
                if (WDebug) {
                    using (StreamWriter sw = new StreamWriter(LogFilePath, true))
                    {
                        sw.WriteLine(DateTime.Now.ToString(c_datetime_format) + " [DEBUG] [TID '" + System.Threading.Thread.CurrentThread.ManagedThreadId + "']: " + sLogLine);
                        if (WGUI)
                            WriteGUI(DateTime.Now.ToString(c_datetime_format) + " [DEBUG] [TID '" + System.Threading.Thread.CurrentThread.ManagedThreadId + "']: " + sLogLine + "\r\n", eType);
                    }
                }
            }
        }

        public void WriteDebug (String sLogLine) {
            WriteDebug(sLogLine, GUIEventType.COMMON);
        }

        public void WriteDebugWarning(String sLogLine)
        {
            WriteDebug(sLogLine, GUIEventType.WARNING);
        }

        public void WriteDebugInformational(String sLogLine)
        {
            WriteDebug(sLogLine, GUIEventType.INFORMATIONAL);
        }

        public void WriteDebugSuccess(String sLogLine)
        {
            WriteDebug(sLogLine, GUIEventType.SUCCESS);
        }

        private void WriteGUI(String sLogLine, GUIEventType eType)
        {
            if (textBox.InvokeRequired)
            {
                AsyncGUIUpdate del = new AsyncGUIUpdate(WriteGUI);
                textBox.Invoke(del, new object[] {sLogLine, eType});
            }
            else {
                switch (eType)
                {
                    case GUIEventType.COMMON:
                        textBox.AppendText(sLogLine);
                        break;
                    case GUIEventType.WARNING:

                        textBox.SelectionStart = textBox.Text.Length;
                        textBox.SelectionBackColor = System.Drawing.Color.Red;
                        textBox.SelectionColor = System.Drawing.Color.White;
                        textBox.AppendText(sLogLine);
                        textBox.SelectionLength = sLogLine.Length;
                        break;
                    case GUIEventType.SUCCESS:
                        textBox.SelectionStart = textBox.Text.Length;
                        textBox.SelectionBackColor = System.Drawing.Color.Green;
                        textBox.SelectionColor = System.Drawing.Color.White;
                        textBox.AppendText(sLogLine);
                        textBox.SelectionLength = sLogLine.Length;
                        break;
                    case GUIEventType.INFORMATIONAL:
                        textBox.SelectionStart = textBox.Text.Length;
                        textBox.SelectionBackColor = System.Drawing.Color.Blue;
                        textBox.SelectionColor = System.Drawing.Color.White;
                        textBox.AppendText(sLogLine);
                        textBox.SelectionLength = sLogLine.Length;
                        break;
                    default:
                        textBox.AppendText(sLogLine);
                        break;
                }
            
            }
        }
     
    }
}

