﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;
using UiO.TRAMP.Logger;
using UiO.TRAMP.Transport;
using System.Net;
using System.Threading;
using System.IO;
using System.Windows.Forms;

namespace UiO.TRAMP.MSRP
{
    class MSRPClass
    {
        private LoggerClass Log;        
        private TCPConnection tcpConn;
        private MSRPSession msrpSession;

        /*Used to decide whether the report sending will be asynchronous or not.
            Consider to have this variable in the GUI so that you can measure difference in performance in this two cases
         */
        private Boolean sendREPORTAsyncBoolean = true; 

        /*
            Used to synchronize sendig of REPORT request such that only one REPORT sending request exists at a time
         *  It is used only when a asynchronous sending of REPORT request is used 
         */
        private ManualResetEvent syncSendingREPROTRequest;

        private ManualResetEvent syncSessionReleaseWithReceivedReports;

        private Stopwatch msrpSw = new Stopwatch();
        private TimeSpan msrpTimeSpan;

        public const string c_default_msrp_port = "2855";
        const int c_msrp_send_chunk_size = 7680;

        private const String c_application_data_file = @"c:\temp\testdoc.pdf";
        private const String c_received_application_base_name = @"c:\temp\receivedDoc";
        private const String c_received_application_extension = ".pdf";

        public Uri localMSRPUri;
        public Uri remoteMSRPUri;

        //this variable will in general be sat by negotiation layer and it represents the result of the precopy algorithm
        public Boolean precopy = true;

        //used for flagging between threads when something goes wrong and someone calls Abort()
        public volatile Boolean aborted = false;

        /*
         * In the sending branch this variable will be used to signal when the final response (REPORT) has been received which will be used to break from the receie loop
           On the other hand do we need to exit this loop at all? It is after all a separate thread that can listen on a socket as long as the application is running 
         */
        public volatile Boolean finalResponseReceived = false;

        // Memory stream that holds the received payload from the remote peer
        public MemoryStream rcvStream = new MemoryStream();

        // Used to control sending loop
        public volatile Boolean sendingLoopStoped = false;

        private Boolean source = false;

        //public Form gui;
        private System.Windows.Forms.ProgressBar progressBar1;

        //delegates for progressBar
        delegate void MaxPBValue(int value);

        delegate void SetPBValue(int value);

        delegate void StepPBValue(int value);

        delegate void InitializePB();

        public MSRPClass (LoggerClass Log, Object gui) { 
            //todo ...
            this.Log = Log;
            this.tcpConn = new TCPConnection(Log);
            this.progressBar1 = (ProgressBar)gui;

            syncSendingREPROTRequest = new ManualResetEvent(true);
            syncSessionReleaseWithReceivedReports = new ManualResetEvent(false);
        }

        /*
            This is a bit of mixing the layers but we need it in order to display progressbar to the user
         */
        private void InitializeProgressBar() {
            /*progressBar1 = new System.Windows.Forms.ProgressBar();

            progressBar1.ForeColor = System.Drawing.Color.LimeGreen;
            progressBar1.Location = new System.Drawing.Point(29, 329);
            progressBar1.Name = "progressBar1";
            progressBar1.Size = new System.Drawing.Size(1177, 23);
            progressBar1.Style = System.Windows.Forms.ProgressBarStyle.Continuous;
            progressBar1.TabIndex = 14;
            progressBar1.Minimum = 1;
            progressBar1.Value = 1;

            gui.Controls.Add(progressBar1);*/

            if (progressBar1.InvokeRequired)
            {
                InitializePB del = new InitializePB(InitializeProgressBar);
                progressBar1.Invoke(del);
            }
            else
            {
                progressBar1.Visible = true;
            }


        }

        public void CalculateLocalMSRPUri(String selectedLocalIPAddress) {            
            String msrpUri = "msrp://" + selectedLocalIPAddress + ":" + c_default_msrp_port + "/" + GetRandomString(10) + ";tcp";
            Log.WriteDebug("UiO.TRAMP.MSRP.CalculateLocalMSRPUri(): Calculated local MSRP URI is: '" + msrpUri + "'.");
            localMSRPUri = new Uri(msrpUri);
        }

        private void Abort() {
            ReleaseResources();
            aborted = true;
            Thread.CurrentThread.Abort();
        }

        public void ReleaseResources() {
            if(tcpConn != null)
                tcpConn.ReleaseResources();
        }

        public void TransferData()
        {
            /*
                Here we will have 2 threds that will deal with:
             * a) sending of data with SEND request  (this will be the main thread)
             * b) receiving of data with REPORT request (this will be the worker thread)
             */

            source = true;

            Log.WriteLog("UiO.TRAMP.MSRP.TransferData(): Transfering data with MSRP protocol.");
            ConnectMSRPSync();
            SendMSRPRequestSEND();

            if (precopy) {
                Log.WriteLog("UiO.TRAMP.MSRP.TransferData(): Waiting for the receiving thread (worker) to receive all reports.");
                syncSessionReleaseWithReceivedReports.WaitOne();
                Log.WriteLog("UiO.TRAMP.MSRP.TransferData(): Receiving thread has finished. Moving on.");
            } else { 
                //Check how to cancel receiving thread when no reports are sent to the source host
            }

            //msrpSw.Stop();

            if (Log.WriteToLog != true) {
                Log.WriteToLog = true;
                msrpTimeSpan = msrpSw.Elapsed;
                Log.WriteLog("UiO.TRAMP.MSRP.TransferData(): The MSRP protocol took '" + msrpTimeSpan.TotalMilliseconds + "' ms.");

                Log.WriteToLog = false;
            }
        }        

        public void ReceiveTransferedData() { 
            /*
                Here we need to receive MSRP SEND requests. This will happen on a main thread
             */

            MemoryStream currentmStream = new MemoryStream();
            msrpSession = new MSRPSession(localMSRPUri, remoteMSRPUri, "");            

            while (!finalResponseReceived) {
                ReceiveMSRPMessageSync(true, currentmStream);
            }

            //Finnaly, when the entire receive round is finished, write the payload to a file
            if (!source) {
                Log.WriteDebug("UiO.TRAMP.MSRP.ReceiveTransferedData(): The local host is a destination host. Saving received file to disk.");
                WritePayloadToFile(rcvStream);
            } else {
                Log.WriteDebug("UiO.TRAMP.MSRP.ReceiveTransferedData(): local host is source host. No file to save.");
            }
        }

        private void SendMSRPRequestSEND() {

            int bytesToBeRead;
            int startRange = 1;
            int endRange = 0;            

            msrpSession = new MSRPSession(localMSRPUri, remoteMSRPUri, GetRandomString(11));

            /*
                We are starting here asynchronous receive operation (on a separate thread) such that we can receive REPORT requests from the peer simultaniously
             */
            ReceiveMSRPMessageAsync();

            try {
                if (File.Exists(c_application_data_file)) {

                    Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): Found application file: '" + c_application_data_file + "'.");
                    
                    using (FileStream fs = new FileStream(c_application_data_file, FileMode.Open, FileAccess.Read)) {

                        //here would be the most appropriate place to start performance measurement of MSRP protocol
                        msrpSw.Start();

                        int payloadLength = (int)fs.Length;
                        Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): Payload length is: '" + payloadLength + "'.");

                        bytesToBeRead = payloadLength;

                        fs.Seek(0, SeekOrigin.Begin);

                        while (bytesToBeRead > 0) {
                            int tmpBufferSize;

                            if (sendingLoopStoped)
                                return;

                            if (bytesToBeRead < c_msrp_send_chunk_size) {
                                tmpBufferSize = bytesToBeRead;
                            } else {
                                tmpBufferSize = c_msrp_send_chunk_size;
                            }

                            //Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): Temporary buffer size is: '" + tmpBufferSize + "'.");
                            byte[] tmpBuffer = new byte[tmpBufferSize];

                            int bytesReadFromStream = fs.Read(tmpBuffer, 0, tmpBufferSize);

                            if (bytesReadFromStream == 0) {
                                //Log.WriteLogWarning("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): Read 0 bytes from the message stream. Aborting.");
                                throw new Exception("No bytes were read from file stream.");
                            } else {
                                //Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): Read '" + bytesReadFromStream + "' B from the file stream.");
                            }

                            String lastLine;

                            endRange = endRange + tmpBufferSize;

                            String msrpSendRequestHeader = CreateMSRPSendRequestHeader(startRange, endRange, payloadLength, out lastLine, msrpSession.messageID);
                            // now we can increment start range for the following chuk
                            startRange = startRange + tmpBufferSize;

                            //Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): Sent following MSRP header in SEND request: \r\n\r\n" + msrpSendRequestHeader);
                            //Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): Last line: '" + lastLine + "'.");
                            //Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): Nubmber of header bytes is: '" + new UTF8Encoding().GetByteCount(msrpSendRequestHeader) + "'.");
                            //Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): Nubmber of lastLine bytes is: '" + new UTF8Encoding().GetByteCount(lastLine) + "'.");

                            byte[] headerBuffer = new UTF8Encoding().GetBytes(msrpSendRequestHeader);
                            byte[] messageWithoutLastLine = new byte[headerBuffer.Length + tmpBufferSize];

                            //merge 
                            Array.Copy(headerBuffer, messageWithoutLastLine, headerBuffer.Length);
                            Array.Copy(tmpBuffer, 0, messageWithoutLastLine, headerBuffer.Length, tmpBufferSize);

                            byte[] lastLineBytes = new UTF8Encoding().GetBytes(lastLine);
                            byte[] message = new byte[messageWithoutLastLine.Length + lastLineBytes.Length];

                            //merge again
                            Array.Copy(messageWithoutLastLine, message, messageWithoutLastLine.Length);
                            Array.Copy(lastLineBytes, 0, message, messageWithoutLastLine.Length, lastLineBytes.Length);

                            if (!aborted) {
                                if (tcpConn.SendSync(message)) {
                                    //Log.WriteLogSuccess("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): The transport layer sent the current message successfully.");
                                } else {
                                    //tcpConn.Disconnect() - shutsdowns and closes socket that failed
                                    //Log.WriteLogWarning("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): Aborting.");
                                    Abort();
                                }
                            } else {
                                //Log.WriteLogWarning("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): Connection to the remote host was aborted because of an error on the receiving thread. Aborting sending (main) thread.");
                                Abort();
                            }

                            bytesToBeRead -= bytesReadFromStream;
                        }
                        msrpSw.Stop();

                    }
                } else {
                    //Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): Unable to find application file: '" + c_application_data_file + "'. Aborting.");
                    Abort();
                }
            } catch (Exception exp) {
                //Log.WriteLogWarning("UiO.TRAMP.MSRP.SendMSRPRequestSEND(): An error occured while extracting application file. " + exp.ToString());
                Abort();
            }            
        }


        private void ReceiveMSRPMessageAsync() {                        
            Thread workerThread = new Thread(new ParameterizedThreadStart(ReceiveAsync));

            workerThread.Start(this);
        }

        private void ReceiveAsync(Object state) {
            /*
               This function waits asynchronously for the requests and responses from the remote peer
             * The idea is to begin this before sending actually begins so that we are ready to receive data when it arrives as a response to the sent data 
             * The receve operation begins on a separate thread
             * Its soul purpose is to update progressbar and thus inform the user about the progress
             */
            MSRPClass msrp = (MSRPClass)state;                                
            MemoryStream currentmStream = new MemoryStream();

            while(!msrp.finalResponseReceived){
                msrp.ReceiveMSRPMessageSync(true, currentmStream);
            }            

            if (precopy){
                msrp.Log.WriteDebug("UiO.TRAMP.MSRP.ReceiveAsync(): signalling the sending thread (main) that all the reports have been received.");
                syncSessionReleaseWithReceivedReports.Set();
                msrp.Log.WriteDebug("UiO.TRAMP.MSRP.ReceiveAsync(): sending thread (main) signalled.");
            }
        }

        private void WritePayloadToFile(MemoryStream rcvStream) {
            Boolean found = false;
            int rcvFileInstance = 0;            

            String fileName = "";

            Log.WriteDebug("UiO.TRAMP.MSRP.WritePayloadToFile(): Writing payload to the file.");

            while (!found) {
                if (File.Exists(c_received_application_base_name + "_" + rcvFileInstance.ToString() + c_received_application_extension)) {
                    Log.WriteDebug("UiO.TRAMP.MSRP.WritePayloadToFile(): File '" + c_received_application_base_name + "_" + rcvFileInstance.ToString() + c_received_application_extension + "' exists. Trying next name.");
                    rcvFileInstance++;
                } else {
                    Log.WriteDebug("UiO.TRAMP.MSRP.WritePayloadToFile(): Writing payload to the file '" + c_received_application_base_name + "_" + rcvFileInstance.ToString() + c_received_application_extension + "'.");
                    found = true;
                    fileName = c_received_application_base_name + "_" + rcvFileInstance.ToString() + c_received_application_extension;
                }
            }

            Log.WriteDebug("UiO.TRAMP.MSRP.WritePayloadToFile(): Writing '" + rcvStream.Length + "' B to the file '" + fileName + "'.");

            rcvStream.Seek(0, SeekOrigin.Begin);

            try {
                //File.WriteAllBytes(fileName, rcvStream.GetBuffer());
                using (FileStream fs = new FileStream(fileName, FileMode.Create, System.IO.FileAccess.Write)) {
                    byte [] streamBytes = new byte[rcvStream.Length];
                    rcvStream.Read(streamBytes, 0, (int)streamBytes.Length);
                    fs.Write(streamBytes, 0, streamBytes.Length);
                }
            } catch (Exception exp) {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.WritePayloadToFile(): An error occured while saving payload to the file '" + fileName + "'. " + exp.ToString());
            }

        }

        private void ReceiveMSRPMessageSync(Boolean doPreparse, MemoryStream currentmStream) {
            int payloadFirstBytePosition = 0;
            /*
                Do not forget to start a timer that will fire after some time and close conenction and thread if not entire message has been received
             * When the message is received the timer will be disposed. The timer will acctually execute at the transport layer
             */

            if (aborted) {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): Connection was disconnected in the sending thread. Aborting receiving thread.");
                Abort();
            }

            while (true) {
                MemoryStream mStream = new MemoryStream();

                if (currentmStream.Length == 0) {
                    Log.WriteLog("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): Processing for empty current Stream.");
                    Log.WriteLog("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): Calling 'Transport' layer to receive data on empty current Stream.");
                    
                    if (tcpConn.ReceiveSync(mStream)) {
                        Log.WriteDebug("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): Transport layer returned '" + mStream.Length + "' B.");

                        try
                        {
                            byte[] buffer = new byte[mStream.Length];

                            mStream.Seek(0, SeekOrigin.Begin); //resetting memory stream position before read opeartion
                            int bytesFromStream = mStream.Read(buffer, 0, buffer.Length);

                            currentmStream.Write(buffer, 0, buffer.Length);

                            //dispose resources
                            buffer = null;
                            mStream.Dispose();

                            //Thread.CurrentThread.Abort();
                            if (doPreparse)
                            {
                                Log.WriteLog("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): Starting MSRPpreparser() function.");
                                payloadFirstBytePosition = MSRPPreParser(currentmStream);
                                Log.WriteDebug("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): Calculated firstByteposition is '" + payloadFirstBytePosition + "'.");
                                if (payloadFirstBytePosition > 0)
                                    break;
                            }
                            else
                            {
                                return; // when we only query for the payload data the outher loop will take care of multiple calls of MSRPReceiveSyn() function
                            }
                        }
                        catch (Exception exp)
                        {
                            Log.WriteLogWarning(exp.ToString());
                        }
                    }
                    else
                    {
                        Log.WriteLogWarning("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): An error occured on transport layer while receiving data. Aborting.");
                        //tcpConn.Disconnect() - shutsdowns and closes socket that failed
                        Abort();
                    }
                } else {
                    Log.WriteLog("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): Processing for NON-empty current Stream.");

                    if (doPreparse) {
                        Log.WriteLog("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): Starting MSRPpreparser() function.");
                        payloadFirstBytePosition = MSRPPreParser(currentmStream);
                        Log.WriteDebug("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): Calculated firstByteposition is '" + payloadFirstBytePosition + "'.");
                        if (payloadFirstBytePosition > 0)
                            break;
                    }

                    Log.WriteLog("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): Calling 'Transport' layer to receive data on NON-empty current Stream.");
                    if (tcpConn.ReceiveSync(mStream)) {
                        Log.WriteDebug("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): Transport layer returned '" + mStream.Length + "' B.");

                        try {

                            byte[] buffer = new byte[mStream.Length];

                            mStream.Seek(0, SeekOrigin.Begin); //resetting memory stream position before read opeartion
                            int bytesFromStream = mStream.Read(buffer, 0, buffer.Length);

                            currentmStream.Write(buffer, 0, buffer.Length);

                            //dispose resources
                            buffer = null;
                            mStream.Dispose();

                        } catch (Exception exp) {
                            Log.WriteLogWarning(exp.ToString());
                        }
                    } else {
                        Log.WriteLogWarning("UiO.TRAMP.MSRP.ReceiveMSRPMessageSync(): An error occured on transport layer while receiving data. Aborting.");
                        //tcpConn.Disconnect() - shutsdowns and closes socket that failed
                        Abort();
                    }

                    if (!doPreparse)
                        return;

                }

            }


            ProcessMSRPHeader(currentmStream, payloadFirstBytePosition);
        }

        private void ProcessMSRPHeader(MemoryStream currentmStream, int payloadPosition) {
            
            int lastHeaderByte = payloadPosition - 3; //-3 to point to the last LF character before CRLF that separates header and payload
            long currentStreamsPosition = currentmStream.Position;

            Log.WriteDebug("UiO.TRAMP.MSRP.ProcessMSRPHeader(): Extracting MSRP header from the input stream. The last header byte is found at position '" + lastHeaderByte + "'.");
            currentmStream.Seek(0, SeekOrigin.Begin);

            byte[] headerBytes = new byte[lastHeaderByte];
            StringBuilder sb = new StringBuilder();

            int bytesRead = currentmStream.Read(headerBytes, 0, headerBytes.Length);
            Log.WriteDebug("UiO.TRAMP.MSRP.ProcessMSRPHeader(): Read '" + bytesRead + "' B from memory stream containing MSRP header.");

            //sb now holds all the received data as UTF8 string
            sb.Append(new UTF8Encoding().GetString(headerBytes));

            Log.WriteDebug("UiO.TRAMP.MSRP.ProcessMSRPHeader(): Received the following header: \r\n\r\n" + sb.ToString());

            String[] headerFields = sb.ToString().Split(new string[] { Environment.NewLine }, StringSplitOptions.None);

            //puting back stream to its current position in case more reads are needed from the transport layer
            currentmStream.Position = currentStreamsPosition;

            if (headerFields[0].Contains("SEND")) {
                //SEND request is received
                Log.WriteLogInformational("UiO.TRAMP.MSRP.ProcessMSRPHeader(): Received MSRP SEND request.");
                MSRPParseSENDRequest(currentmStream, payloadPosition, headerFields);
            } else if (headerFields[0].Contains("REPORT")){
                //REPORT request is received
                Log.WriteLogInformational("UiO.TRAMP.MSRP.ProcessMSRPHeader(): Received MSRP REPORT request.");
                MSRPParseREPORTRequest(currentmStream, payloadPosition, headerFields);
            } else if (headerFields[0].Contains("400")) {
                //400 response is received
                Log.WriteLogInformational("UiO.TRAMP.MSRP.ProcessMSRPHeader(): Received 400 MSRP response.");
                MSRPParse400Response(headerFields);
            } else { 
                // nothing else is supported so this situation is ilegal
                Log.WriteLogWarning("UiO.TRAMP.MSRP.ProcessMSRPHeader(): Received unsupported MSRP message. Aborting.");
                Abort();
            }
        }

        private void MSRPParseSENDRequest(MemoryStream currentmStream, int payloadPosition, String [] headerFields) {

            String rangeStartString;
            String rangeEndString;
            String totalSizeString;

            int rangeStart = 0;
            int rangeEnd = 0;
            int totalSize = 0;

            Boolean hFound = false;

            Boolean isRequestUnintelligible = false;

            String requestLineTransactionID = headerFields[0].Split(' ')[1].Trim();

            foreach (String header in headerFields) {
                if (header.StartsWith("Byte-Range"))
                {
                    Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Found 'Byte-Range' header field: '" + header + "'.");
                    hFound = true;

                    rangeStartString = header.Split(':')[1].Trim().Split('-')[0].Trim();
                    Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Found 'range start' with value of '" + rangeStartString + "'.");

                    rangeEndString = header.Split(':')[1].Trim().Split('-')[1].Trim().Split('/')[0].Trim();
                    Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Found 'range end' with value of '" + rangeEndString + "'.");

                    totalSizeString = header.Split(':')[1].Trim().Split('-')[1].Trim().Split('/')[1].Trim();
                    Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Found 'total size' with value of '" + totalSizeString + "'.");

                    if (int.TryParse(rangeStartString, out rangeStart))
                    {
                        Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Extracted 'range start'.");
                    }
                    else
                    {
                        Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Unable to extract 'range start'. Aborting.");
                        //Here we will be forced to send a response of type 4xx to signal the other side that something went wrong. For now it is just an abort.
                        isRequestUnintelligible = true;
                        break;
                    }

                    if (int.TryParse(rangeEndString, out rangeEnd))
                    {
                        Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Extracted 'range end'.");
                    }
                    else
                    {
                        Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Unable to extract 'range end'. Aborting.");
                        //Here we will be forced to send a response of type 4xx to signal the other side that something went wrong. For now it is just an abort.
                        isRequestUnintelligible = true;
                        break;
                    }

                    if (int.TryParse(totalSizeString, out totalSize))
                    {
                        Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Extracted 'total size'.");
                    }
                    else
                    {
                        Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Unable to extract 'range start'. Aborting.");
                        //Here we will be forced to send a response of type 4xx to signal the other side that something went wrong. For now it is just an abort.
                        isRequestUnintelligible = true;
                        break;
                    }
                }
            }

            if (!hFound) {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Unable to find 'Byte-Range' header field in input stream. Aborting.");
                isRequestUnintelligible = true;
            }

            if (isRequestUnintelligible) {
                SendMSRP400ResponseSync(requestLineTransactionID);
                Abort();
            }


            String toUriString = getToLine(headerFields).Split(' ')[1].Trim();
            String fromUriString = getFromLine(headerFields).Split(' ')[1].Trim();

            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Received 'ToUri' is: '" + toUriString + "'.");
            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Received 'FromUri' is: '" + fromUriString + "'.");


            Uri toUri = new Uri(toUriString);
            Uri fromUri = new Uri(fromUriString);

            String messageID = getMessageIDLine(headerFields).Split(':')[1].Trim();
            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Received 'messageID' is: '" + messageID + "'.");

            String successReportValue = getSuccessReportLine(headerFields).Split(':')[1].Trim();
            String failureReportValue = getFailureReportLine(headerFields).Split(':')[1].Trim();

            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Retreived success-report value of '" + successReportValue + "'.");
            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Retreived failure-report value of '" + failureReportValue + "'.");

            // Compare the received REPORT request with the current MSRP session

            if (msrpSession.messageID == String.Empty) {
                // the first SEND request from the remote peer
                Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): This is the first received SEND request. Storing messageID '" + messageID + "'.");
                msrpSession.messageID = messageID;
            }

            int contentLength = rangeEnd - rangeStart; // the current payload size 
            int lastPayloadBytePosition = payloadPosition + contentLength;

            currentmStream.Position = currentmStream.Length; // the next byte in the stream to write to

            if ((msrpSession.localUri.Equals(toUri)) && (msrpSession.remoteUri.Equals(fromUri)) && (msrpSession.messageID.Equals(messageID, StringComparison.Ordinal))) {
                Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Received request matches the MSRP session on the local host.");
            } else {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Received request does not match MSRP session on the local host. Ignoring.");                
                RewindMemoryStream(currentmStream); 
                return;
            }                                    

            /*
                With this loop we are trying to make sure that the entire payload for this SEND request has been received
             */

            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Comparing '" + currentmStream.Length + "' <= '" + lastPayloadBytePosition + "'.");

            if (contentLength > 0) {
                if ((currentmStream.Length) <= (payloadPosition + contentLength)) {
                    while ((currentmStream.Length - 1 - payloadPosition) < contentLength) {
                        // in the worst case scenarion this potentially infinite loop should be interrupted by a transaction timer                   
                        Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Calculated length is: '" + (currentmStream.Length - 1 - payloadPosition) + "' B.");
                        Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Length extracted from Content-Length header field is: " + contentLength + "' B.");
                        Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Not all payload data has been received. Returning to socket for more data.");
                        ReceiveMSRPMessageSync(false, currentmStream);
                    }
                } else {
                    Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Current stream contains entire message. No need to check for more data on the transport layer.");
                }
            } else {
                Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Received request without the payload (Content-Length = 0). No need to check for more data on the transport layer.");
            }

            int currentmStreamPosition = (int)currentmStream.Position;  // the stream end                      

            int lastTransactionIDDelimiterBytePosition = 0;

            while (true) {
                currentmStream.Seek(lastPayloadBytePosition + 1, SeekOrigin.Begin); // points to the last byte of the payload + 1 = first last line byte ('-')
                lastTransactionIDDelimiterBytePosition = FindTransactionIDFirstBytePosition(currentmStream);
                if (lastTransactionIDDelimiterBytePosition > 0) // if > 0 points to the first byte of TransactionID
                    break;
                currentmStream.Position = (long)currentmStreamPosition;
                ReceiveMSRPMessageSync(false, currentmStream);
                currentmStreamPosition = (int)currentmStream.Position;
            }

            //stream now points to last '-' char in MSRP message delimiter, right before transactionID

            int msrpContnuationFlagPosition = 0;

            while (true) {
                currentmStream.Seek(lastTransactionIDDelimiterBytePosition, SeekOrigin.Begin);
                msrpContnuationFlagPosition = FindcontinuationFlagBytePosition(currentmStream);
                if (msrpContnuationFlagPosition > 0) // if > 0 points to the first byte after continuation flag
                    break;
                currentmStream.Position = (long)currentmStreamPosition;
                ReceiveMSRPMessageSync(false, currentmStream);
                currentmStreamPosition = (int)currentmStream.Position;
            }

            //stream now point to the first byte after the continuation flag

            /*
                At this point we have:
             * a) input stream
             * b) position of the payload
             * c) position of the end pf the payload
             * d) position of the transactionID 
             * in the stream
             */

            // extract payload
            byte[] rcvTmpBuffer = new byte[contentLength + 1];
            currentmStreamPosition = (int)currentmStream.Position; // the first byte after continuation flag

            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Positioning the current memory stream to the first byte of the payload '" + payloadPosition + "'.");
            currentmStream.Seek(payloadPosition, SeekOrigin.Begin);
            int readBytes = currentmStream.Read(rcvTmpBuffer, 0, rcvTmpBuffer.Length);

            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Read '" + readBytes + "' B of payload from the current stream.");

            rcvStream.Write(rcvTmpBuffer, 0, rcvTmpBuffer.Length);

            // extract transactionID
            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Positioning the current memory stream to the first byte of the transactionID '" + (lastTransactionIDDelimiterBytePosition) + "'.");
            currentmStream.Seek(lastTransactionIDDelimiterBytePosition, SeekOrigin.Begin);

            rcvTmpBuffer = new byte[msrpContnuationFlagPosition - lastTransactionIDDelimiterBytePosition - 1]; // -1 to exclude continuation flag

            readBytes = currentmStream.Read(rcvTmpBuffer, 0, rcvTmpBuffer.Length);
            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Read '" + readBytes + "' B of payload from the current stream.");

            String lastLineTransactionID = new UTF8Encoding().GetString(rcvTmpBuffer);

            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Read transactionID '" + lastLineTransactionID + "' from the last line.");
            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Comparing two transactioID's: '" + requestLineTransactionID + "' <-> '" + lastLineTransactionID + "'.");

            if (!lastLineTransactionID.Equals(requestLineTransactionID)) {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): TransactionID from the request line and the last line are not the same. Sending error response.");
                SendMSRP400ResponseSync(requestLineTransactionID);
                finalResponseReceived = true;
                return;
            }

            // send REPORT request
            currentmStream.Seek(msrpContnuationFlagPosition, SeekOrigin.Begin); // this is actually the first byte of the next message            

            if (successReportValue.Equals("yes", StringComparison.OrdinalIgnoreCase))
            {
                if (!sendREPORTAsyncBoolean)
                {
                    Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): The sender has requested REPORT requests. Sending synchronously.");
                    SendMSRPRequestREPORT(rangeStart, rangeEnd, totalSize);
                }
                else
                {
                    Log.WriteLog("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): The sender has requested REPORT requests. Sending asynchronously.");
                    if (!aborted)
                    {
                        Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Waiting for the sending of the previous REPORT request to finish.");
                        syncSendingREPROTRequest.WaitOne();
                    }
                    else
                    {
                        Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Sending thread was aborted due to fail. Aborting receiving (main) thread as well.");
                        Abort();
                    }

                    Log.WriteLog("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Resetting synchronization primitive for the next request.");
                    syncSendingREPROTRequest.Reset();
                    Log.WriteLog("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Synchronization primitive reset.");
                    SendMSRPRequestREPORTAsync(rangeStart, rangeEnd, totalSize);
                }
            }
            else
            {
                Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): The sender has not requested REPORT requests.");
            }

            if (rangeEnd == totalSize) {
                Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseSENDRequest(): Entire message has been received. Exiting receive loop.");
                finalResponseReceived = true;
                return;
            }

            //finally rewind the currentStream such that its first position points to the first byte in the rest of the payload
            RewindMemoryStream(currentmStream);
            

        }

        private void SendMSRPRequestREPORTAsync(int rStart, int rEnd, int tSize) {
            MSRPReportState state = new MSRPReportState(rStart, rEnd, tSize);

            Thread reportSendThread = new Thread(new ParameterizedThreadStart(SendReportAsync)); ;
            reportSendThread.Start(state);
        }

        private void SendReportAsync(Object state) {
            MSRPReportState rState = (MSRPReportState)state;

            Log.WriteLog("UiO.TRAMP.MSRP.SendReportAsync(): Sending MSRP REPORT request asynchronously.");
            SendMSRPRequestREPORT(rState.rStart, rState.rEnd, rState.tSize);
            Log.WriteLog("UiO.TRAMP.MSRP.SendReportAsync(): Signalling the main thread that the REPORT request has been sent. The new request can be sent down the pipeline.");
            syncSendingREPROTRequest.Set();
        }

        private void SendMSRP400ResponseSync(String transactionID) { 
            // always executes synchronously since the thread aborts after it anyhow

            int bytesToBeRead;
            MemoryStream ms = new MemoryStream();

            try
            {
                Log.WriteLog("UiO.TRAMP.MSRP.SendMSRP400ResponseSync(): Sending 400 response.");

                StringBuilder sb = new StringBuilder();

                sb.Append("MSRP " + transactionID + "400 Request was unintelligible");
                sb.Append("\r\n");
                sb.Append("To-Path: " + msrpSession.remoteUri.ToString());
                sb.Append("\r\n");
                sb.Append("From-Path: " + msrpSession.localUri.ToString());
                sb.Append("\r\n");
                sb.Append("-------" + transactionID + "$");
                sb.Append("\r\n");
                sb.Append("\r\n"); //(kremt)

                byte[] payload = new UTF8Encoding().GetBytes(sb.ToString());

                Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRP400ResponseSync(): Sending the following 400 response: \r\n\r\n" + sb.ToString());

                int payloadLength = payload.Length;
                Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRP400ResponseSync(): Payload length is: '" + payloadLength + "'.");

                bytesToBeRead = payloadLength;

                ms.Write(payload, 0, payloadLength);
                ms.Seek(0, SeekOrigin.Begin);

                while (bytesToBeRead > 0)
                {
                    int tmpBufferSize;

                    if (bytesToBeRead < c_msrp_send_chunk_size)
                    {
                        tmpBufferSize = bytesToBeRead;
                    }
                    else
                    {
                        tmpBufferSize = c_msrp_send_chunk_size;
                    }

                    Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRP400ResponseSync(): Temporary buffer size is: '" + tmpBufferSize + "'.");
                    byte[] tmpBuffer = new byte[tmpBufferSize];

                    int bytesReadFromStream = ms.Read(tmpBuffer, 0, tmpBufferSize);

                    if (bytesReadFromStream == 0)
                    {
                        Log.WriteLog("UiO.TRAMP.MSRP.SendMSRP400ResponseSync(): Read 0 bytes from the message stream. Aborting.");
                        throw new Exception("No bytes were read from file stream.");
                    }

                    if (tcpConn.SendSync(tmpBuffer))
                    {
                        Log.WriteLogSuccess("UiO.TRAMP.MSRP.SendMSRP400ResponseSync(): The transport layer sent the current message successfully.");
                    }
                    else
                    {
                        //tcpConn.Disconnect() - shutsdowns and closes socket that failed
                        Log.WriteLogWarning("UiO.TRAMP.MSRP.SendMSRP400ResponseSync(): Aborting.");                                                
                    }

                    bytesToBeRead -= bytesReadFromStream;
                }
            } catch (Exception exp) {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.SendMSRP400ResponseSync(): An error occured while sednding 400 response. " + exp.ToString());                                
            }
        }

        private void SendMSRPRequestREPORT(int sRange, int eRange, int tSize) {
            /*
                This function is a synchronous version of MSRP REPORT request SEND function
             *  It could be realized as asynchronous as well, but in that case we would need a synchronization between 
             *  sending threads since we cannot have two threads sending to the same remote end point. The value of asynchronous sending 
             *  of REPORT requests will be closer investigated during the evaluation
             */

            int bytesToBeRead;
            MemoryStream ms = new MemoryStream();                                    

            try
            {
                Log.WriteLog("UiO.TRAMP.MSRP.SendMSRPRequestREPORT(): Sending MSRP REPORT request.");

                    StringBuilder sb = new StringBuilder();
                    String newTransactionID = GetRandomString(10);
                
                    sb.Append("MSRP " + newTransactionID + " REPORT");
                    sb.Append("\r\n");
                    sb.Append("To-Path: " + msrpSession.remoteUri.ToString());
                    sb.Append("\r\n");
                    sb.Append("From-Path: " + msrpSession.localUri.ToString());
                    sb.Append("\r\n");
                    sb.Append("Message-ID: " + msrpSession.messageID);
                    sb.Append("\r\n");
                    sb.Append("Byte-Range: " + sRange + "-" + eRange + "/" + tSize);
                    sb.Append("\r\n");
                    sb.Append("Status: 000" + " 200 OK");
                    sb.Append("\r\n");
                    sb.Append("-------" + newTransactionID + "$");
                    sb.Append("\r\n");
                    sb.Append("\r\n"); //(kremt)

                    byte[] payload = new UTF8Encoding().GetBytes(sb.ToString());

                    Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestREPORT(): Sending the following REPORT: \r\n\r\n" + sb.ToString());
                    
                    int payloadLength = payload.Length;
                    Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestREPORT(): Payload length is: '" + payloadLength + "' B.");

                    bytesToBeRead = payloadLength;

                    ms.Write(payload, 0, payloadLength);
                    ms.Seek(0, SeekOrigin.Begin);  

                    while (bytesToBeRead > 0)
                    {
                        int tmpBufferSize;

                        if (sendingLoopStoped)
                            return;

                        if (bytesToBeRead < c_msrp_send_chunk_size) {
                            tmpBufferSize = bytesToBeRead;
                        } else {
                            tmpBufferSize = c_msrp_send_chunk_size;
                        }

                        Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestREPORT(): Temporary buffer size is: '" + tmpBufferSize + "' B.");
                        byte[] tmpBuffer = new byte[tmpBufferSize];

                        int bytesReadFromStream = ms.Read(tmpBuffer, 0, tmpBufferSize);

                        if (bytesReadFromStream == 0) {
                            Log.WriteLog("UiO.TRAMP.MSRP.SendMSRPRequestREPORT(): Read 0 bytes from the message stream. Aborting.");
                            throw new Exception("No bytes were read from file stream.");
                        }

                        if(!aborted){
                            if (tcpConn.SendSync(tmpBuffer)) {
                                Log.WriteLogSuccess("UiO.TRAMP.MSRP.SendMSRPRequestREPORT(): The transport layer sent the current message successfully.");
                            } else {
                                //tcpConn.Disconnect() - shutsdowns and closes socket that failed
                                Log.WriteLogWarning("UiO.TRAMP.MSRP.SendMSRPRequestREPORT(): Aborting.");
                                finalResponseReceived = true; // abort receiving
                                Abort();
                            }
                        } else {
                            Log.WriteLogWarning("UiO.TRAMP.MSRP.SendMSRPRequestREPORT(): Connection to the remote host was aborted because of an error on the receiving thread. Aborting sending (worker) thread.");
                            Abort();
                        }

                        bytesToBeRead -= bytesReadFromStream;
                    }                            
            } catch (Exception exp) {
                Log.WriteDebug("UiO.TRAMP.MSRP.SendMSRPRequestREPORT(): An error occured while sending MSRP REPORT request. " + exp.ToString());
                finalResponseReceived = true; // abort receiving
                Abort();
            }
        }

        private int FindcontinuationFlagBytePosition(MemoryStream currentmStream) {
            int position = 0;

            if (currentmStream.Length == 0) {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.FindcontinuationFlagBytePosition(): Found no data in the receive buffer. Aborting.");
                //tcpConn.Disconnect() - shutsdowns and closes socket that failed
                Abort();
            } else {
                Log.WriteDebug("UiO.TRAMP.MSRP.FindcontinuationFlagBytePosition(): Current stream position is '" + currentmStream.Position + "'.");
            }
           

            while (HasMoreData(currentmStream)) {
                char ch = Convert.ToChar(currentmStream.ReadByte());

                if ((ch == '$') || (ch == '+')) {
                    //stream now points to the first byte after the continuation flag
                    position = (int)currentmStream.Position;
                    Log.WriteDebug("UiO.TRAMP.MSRP.FindcontinuationFlagBytePosition(): Found  continuation flag '" + ch.ToString() +  "' at position '" + position + "' in the stream.");

                    return position;
                }                
            }

            return -1;
        }

        private int FindTransactionIDFirstBytePosition(MemoryStream currentmStream) { 
            int position = 0;                 

            if (currentmStream.Length == 0) {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.FindTransactionIDFirstBytePosition(): Found no data in the receive buffer. Aborting.");
                //tcpConn.Disconnect() - shutsdowns and closes socket that failed
                Abort();
            } else {
                Log.WriteDebug("UiO.TRAMP.MSRP.FindTransactionIDFirstBytePosition(): Current stream position is '" + currentmStream.Position + "'.");
            }            

            while (HasMoreData(currentmStream))
            {
                char ch = Convert.ToChar(currentmStream.ReadByte());
                //Log.WriteDebug("Processing Char '" + ch + "'");
                switch (position)
                {
                    case 0:
                        if (ch == '-') { position++; }
                        break;
                    case 1:
                        if (ch == '-') { position++; }
                        else
                        {
                            //ResetData(currentmStream, ref position);
                            position = 0; 
                        }
                        break;
                    case 2:
                        if (ch == '-') { position++; }
                        else
                        {
                            //ResetData(currentmStream, ref position);
                            position = 0;
                        }
                        break;
                    case 3:
                        if (ch == '-') { position++; }
                        else
                        {
                            //ResetData(currentmStream, ref position);
                            position = 0;
                        }
                        break;
                    case 4:
                        if (ch == '-') { position++; }
                        else
                        {
                            //ResetData(currentmStream, ref position);
                            position = 0;
                        }
                        break;
                    case 5:
                        if (ch == '-') { position++; }
                        else
                        {
                            //ResetData(currentmStream, ref position);
                            position = 0;
                        }
                        break;
                    case 6:
                        if (ch == '-') { 
                            // currentStream.Position holds the first byte of the transactionID
                            position = (int)currentmStream.Position;
                            Log.WriteDebug("UiO.TRAMP.MSRP.FindTransactionIDFirstBytePosition(): Found 7 '-' message delimiters. The transactionID begins at position '" + position + "' in the stream.");

                            return position;
                        } else {
                            position = 0;
                        }
                        break;                    
                    default:
                        break;
                }
            }            
            return -1;
        }

        private void MSRPParseREPORTRequest(MemoryStream currentmStream, int payloadPosition, String [] headerFields) {
            /*
                The soul purpose of this function is to receive the REPORT request and then to display the progress to the 
             * user in GUI
             */            

            String rangeStartString;
            String rangeEndString;
            String totalSizeString;

            int rangeStart = 0;
            int rangeEnd = 0;
            int totalSize = 0; 


            Boolean hFound = false;
            //Since this is the REPORT request with no body we know that the entire request is received when the double 'newline' is detected
            foreach (String header in headerFields) {
                if (header.StartsWith("Byte-Range")) {
                    Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Found 'Byte-Range' header field: '" + header + "'.");
                    hFound = true;

                    rangeStartString = header.Split(':')[1].Trim().Split('-')[0].Trim();
                    Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Found 'range start' with value of '" + rangeStartString + "'.");

                    rangeEndString = header.Split(':')[1].Trim().Split('-')[1].Trim().Split('/')[0].Trim();
                    Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Found 'range end' with value of '" + rangeEndString + "'.");

                    totalSizeString = header.Split(':')[1].Trim().Split('-')[1].Trim().Split('/')[1].Trim();
                    Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Found 'total size' with value of '" + totalSizeString + "'.");

                    if (int.TryParse(rangeStartString, out rangeStart)) {
                        Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Extracted 'range start'.");
                    } else {
                        Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Unable to extract 'range start'. Aborting.");
                        //Here we will be forced to send a response of type 4xx to signal the other side that something went wrong. For now it is just an abort.
                        Abort();
                    }

                    if (int.TryParse(rangeEndString, out rangeEnd)) {
                        Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Extracted 'range end'.");
                    } else {
                        Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Unable to extract 'range end'. Aborting.");
                        //Here we will be forced to send a response of type 4xx to signal the other side that something went wrong. For now it is just an abort.
                        Abort();
                    }

                    if (int.TryParse(totalSizeString, out totalSize)) {
                        Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Extracted 'total size'.");
                    } else {
                        Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Unable to extract 'range start'. Aborting.");
                        //Here we will be forced to send a response of type 4xx to signal the other side that something went wrong. For now it is just an abort.
                        Abort();
                    }
                }
            }
            
            if (!hFound) {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Unable to find 'Byte-Range' header field in input stream. Aborting.");
                Abort();
            }

            if (rangeStart == 1) { // the first REPORT request to be received, because we only need to set this value once
                InitializeProgressBar();
                SetPBMaximum(totalSize);
            }

            String toUriString = getToLine(headerFields).Split(' ')[1].Trim();
            String fromUriString = getFromLine(headerFields).Split(' ')[1].Trim();

            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Received 'ToUri' is: '" + toUriString + "'.");
            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Received 'FromUri' is: '" + fromUriString + "'.");


            Uri toUri = new Uri(toUriString);
            Uri fromUri = new Uri(fromUriString);

            String messageID = getMessageIDLine(headerFields).Split(':')[1].Trim();
            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Received 'messageID' is: '" + messageID + "'.");


            // Compare the received REPORT request with the current MSRP session

            if ((msrpSession.localUri.Equals(toUri)) && (msrpSession.remoteUri.Equals(fromUri)) && (msrpSession.messageID.Equals(messageID, StringComparison.Ordinal))) {
                Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Received request matches the MSRP session on the local host.");
            } else {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Received request does not match MSRP session on the local host. Ignoring.");
                return;
            }

            //purge the transaction state on the local host

            if (msrpSession.transactionStates.ContainsKey(rangeStart + "-" + rangeEnd)) {
                Log.WriteLog("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Removing " + rangeStart + "-" + rangeEnd + "<->" + msrpSession.transactionStates[rangeStart + "-" + rangeEnd] + " from local state.");
                msrpSession.transactionStates.Remove(rangeStart + "-" + rangeEnd);
            } else { 
                Log.WriteLog("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Unable to find '" + rangeStart + "-" + rangeEnd + "' state");
            }

            if (rangeEnd == totalSize) {
                Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Entire message has been received. Exiting receive loop.");
                finalResponseReceived = true;
                //progressBar1.Value = progressBar1.Maximum;
                SetProgressBarValue(totalSize);
                //progressBar1.PerformStep();                
            } else {
                int step = rangeEnd - rangeStart + 1;
                Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParseREPORTRequest(): Incrementing progressBar with the '" + step + "' units.");
                //progressBar1.Step = step;
                //progressBar1.PerformStep();
                StepProgressBarValue(step);
            }

            //finally rewind the currentStream such that its first position points to the first byte in the rest of the payload
            RewindMemoryStreamForReportRequest(currentmStream, payloadPosition);
        }

        private void RewindMemoryStreamForReportRequest(MemoryStream currentmStream, int payloadPosition)
        {
            if (currentmStream.Length == payloadPosition) {
                Log.WriteDebug("UiO.TRAMP.MSRP.RewindMemoryStreamForReportRequest(): The current MSRP REPORT request is the last one in the stream. Resetting the entire stream to '0'.");
                currentmStream.SetLength(0);
            } else {
                Log.WriteDebug("UiO.TRAMP.MSRP.RewindMemoryStreamForReportRequest(): There is more data in the current stream containing MSRP REPORTS. Resetting to the next MSRP REPORT in the stream.");
                int tmpBufferSize = (int)(currentmStream.Length - payloadPosition);
                byte[] tmpBuffer = new byte[tmpBufferSize];

                currentmStream.Seek(payloadPosition, SeekOrigin.Begin);
                currentmStream.Read(tmpBuffer, 0, tmpBufferSize);

                currentmStream.SetLength(tmpBufferSize);
                currentmStream.Seek(0, SeekOrigin.Begin);
                currentmStream.Write(tmpBuffer, 0, tmpBufferSize);
            }
        }

        private void SetPBMaximum(int max) {
            if (progressBar1.InvokeRequired)
            {
                MaxPBValue del = new MaxPBValue(SetPBMaximum);
                progressBar1.Invoke(del, new object[] { max });
            }
            else
            {
                progressBar1.Maximum = max;
            }
        }

        private void SetProgressBarValue(int value) { 
            

            if (progressBar1.InvokeRequired)
            {
                SetPBValue del = new SetPBValue(SetProgressBarValue);
                progressBar1.Invoke(del, new object[] { value });
            }
            else
            {
                progressBar1.Value = value;
            }
        }

        private void StepProgressBarValue(int value)
        {


            if (progressBar1.InvokeRequired)
            {
                StepPBValue del = new StepPBValue(StepProgressBarValue);
                progressBar1.Invoke(del, new object[] { value });
            }
            else
            {                
                progressBar1.Step = value;
                progressBar1.PerformStep();
            }
        }

        private void RewindMemoryStream(MemoryStream currentmStream) {
            if (currentmStream.Length == (currentmStream.Position)) {
                Log.WriteDebug("UiO.TRAMP.MSRP.RewindMemoryStream(): The current MSRP message is the last one in the stream. Resetting.");
                currentmStream.SetLength(0);
            } else {
                Log.WriteDebug("UiO.TRAMP.MSRP.RewindMemoryStream(): There is more data in the current stream. Resetting.");
                int tmpBufferSize = (int)(currentmStream.Length - currentmStream.Position);
                byte[] tmpBuffer = new byte[tmpBufferSize];

                currentmStream.Read(tmpBuffer, 0, tmpBufferSize);
                
                currentmStream.SetLength(tmpBufferSize);
                currentmStream.Seek(0, SeekOrigin.Begin);
                currentmStream.Write(tmpBuffer, 0 , tmpBufferSize);
            }
        }

        private void MSRPParse400Response(String [] headerFields) {
            /*
                The purpose of this function is to :
             * a) abort the sending loop in the sending thread
             * b) abort the receiving loop and inform the user
             * 
             * We say that here we will only support one negative response and that is 400. In return it will be used for everyting that goes wrong, 
             * but according to design there is not much that could go wrong here
             */
            Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParse400Response(): Received 400 error response. Aborting both sending and receiving loops.");

            String transactionID = headerFields[0].Split(' ')[1].Trim();
            String responseComment = headerFields[0].Split(' ')[3].Trim();

            Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParse400Response(): " + responseComment);

            if (msrpSession.transactionStates.ContainsValue(transactionID)) {
                Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParse400Response(): Found matching transactionID on the local host '" + transactionID + "'.");
            } else {
                Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParse400Response(): Unable to find matching transaction '" + transactionID + "' on the local host. Ignoring.");
                return;
            }

            String toUriString = getToLine(headerFields).Split(' ')[1].Trim();
            String fromUriString = getFromLine(headerFields).Split(' ')[1].Trim();

            Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParse400Response(): Received 'ToUri' is: '" + toUriString + "'.");
            Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParse400Response(): Received 'FromUri' is: '" + fromUriString + "'.");


            Uri toUri = new Uri(toUriString);
            Uri fromUri = new Uri(fromUriString);

            // Compare the received REPORT request with the current MSRP session
            if ((msrpSession.localUri.Equals(toUri)) && (msrpSession.remoteUri.Equals(fromUri))) {
                Log.WriteDebug("UiO.TRAMP.MSRP.MSRPParse400Response(): Received request matches the MSRP session on the local host.");
            } else {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPParse400Response(): Received request does not match MSRP session on the local host. Ignoring.");
                return;
            }


            //stoping receive loop
            finalResponseReceived = true;

            //stoping send loop
            sendingLoopStoped = true;

        }

        private int MSRPPreParser(MemoryStream currentmStream) {

            int headerPosition = 0; // first byte of the payload or +1 byte of the header position in stream
            int streamsCurrentPosition = 0;            

            if (currentmStream.Length == 0) {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.MSRPPreParser(): Found no data in the receive buffer. Aborting.");
                //tcpConn.Disconnect() - shutsdowns and closes socket that failed
                Abort();
            } else {
                streamsCurrentPosition = (int)currentmStream.Position;
                Log.WriteDebug("UiO.TRAMP.MSRP.MSRPPreParser(): Current stream position is '" + streamsCurrentPosition + "'.");
            }

            currentmStream.Seek(0, SeekOrigin.Begin); //resetting current streams' position so that it could be analysed

            while (HasMoreData(currentmStream)) {
                char ch = Convert.ToChar(currentmStream.ReadByte());               

                switch (headerPosition)
                {
                    case 0:
                        if (ch == '\r') { headerPosition++; }
                        break;
                    case 1:
                        if (ch == '\n') { headerPosition++; }
                        else
                        {
                            ResetData(currentmStream, ref headerPosition);
                        }
                        break;
                    case 2:
                        if (ch == '\r') { headerPosition++; }
                        else
                        {
                            ResetData(currentmStream, ref headerPosition);
                        }
                        break;
                    case 3:
                        if (ch == '\n')
                        {
                            // currentStream.Position holds the first byte of the payload, if any
                            headerPosition = (int)currentmStream.Position;                            
                            // the stream points where it did before the function started. This is neccasary in case MSRPParse function must receive more data from transport layer
                            currentmStream.Position = (long)streamsCurrentPosition;
                            Log.WriteDebug("UiO.TRAMP.MSRP.MSRPPreParser(): Found double 'new line' in the input stream. The payload begins at position '" + headerPosition + "' in the stream.");

                            return headerPosition;
                        }
                        else
                        {
                            ResetData(currentmStream, ref headerPosition);
                        }
                        break;
                    default:
                        break;
                }                

            }

            /*
             * If we came so far the two only possibilities are that not enough data has been read to encompas entire header
             * or that the request is not formed correctly (meaning no double new-line is found in the received string). 
             * The latter wiill not be considered, at least not for now, but the former can be remedied with extra read from transport layer
             */
            currentmStream.Position = (long)streamsCurrentPosition;
            return -1;
        }

        private Boolean HasMoreData(MemoryStream mStream)
        {
            return (int)(mStream.Length - mStream.Position) > 0;
        }

        private void ResetData(MemoryStream mStream, ref int postition)
        {
            postition = 0;

            int currentPosition = (int)mStream.Position;
            if (currentPosition != 0) { mStream.Position = (long)(currentPosition - 1); }
        }

        private String CreateMSRPSendRequestHeader(int startRange, int endRange, int messageSize, out String lastLine, String messageID) {
            StringBuilder sb = new StringBuilder();

            String transactionID = GetRandomString(8);
            String successRaport = "no";

            String key = startRange.ToString() + "-" + endRange.ToString();

            /*
                Need this one to keep the transaction states of the outgoing request so that I can match them in case 400 response returns
             * They will be purged as the appropriate REPORT request are received ... what about SEND where Success-Report = no?
             */

            Log.WriteLog("UiO.TRAMP.MSRP.CreateMSRPSendRequestHeader(): Adding key-value pair: '" + key + " <-> " + transactionID + "'.");

            if (!(msrpSession.transactionStates.ContainsKey(key))) {                
                msrpSession.transactionStates.Add(startRange.ToString() + "-" + endRange.ToString(), transactionID);
                Log.WriteLog("UiO.TRAMP.MSRP.CreateMSRPSendRequestHeader(): Added.");
            } else {
                Log.WriteLog("UiO.TRAMP.MSRP.CreateMSRPSendRequestHeader(): Key '" + key + "' already exists.");                
            }

            if (endRange == messageSize) {
                lastLine = "-------" + transactionID + "$";
            } else {
                lastLine = "-------" + transactionID + "+";
            }

            //lastLine = lastLine + "\r\n";

            if (precopy)
                successRaport = "yes";
            

            sb.Append("MSRP " + transactionID + " SEND");
            sb.Append("\r\n");
            sb.Append("To-Path: " + remoteMSRPUri.ToString());
            sb.Append("\r\n");
            sb.Append("From-Path: " + localMSRPUri.ToString());
            sb.Append("\r\n");
            sb.Append("Message-ID: " + messageID);
            sb.Append("\r\n");
            sb.Append("Byte-Range: " + startRange + "-" + endRange + "/" + messageSize);
            sb.Append("\r\n");
            sb.Append("Success-Report: " + successRaport);
            sb.Append("\r\n");
            sb.Append("Failure-Report: partial");
            sb.Append("\r\n");
            sb.Append("Content-Type: application/octet-stream");
            sb.Append("\r\n");
            sb.Append("\r\n");

            return sb.ToString();
        }        

        private void ConnectMSRPSync() {
            IPAddress ipAddress = null;

            Log.WriteLog("UiO.TRAMP.MSRP.ConnectMSRPSync(): Establishing connection to the remote host.");

            try {
                ipAddress = IPAddress.Parse(remoteMSRPUri.DnsSafeHost);
            } catch (Exception exp) {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.ConnectMSRPSync(): An error occured while parsing remote IPAddress. Aborting.");
                Log.WriteLogWarning("UiO.TRAMP.MSRP.ConnectMSRPSync():" + exp.ToString());
                Abort();
            }

            if (tcpConn.ConnectSync(ipAddress, remoteMSRPUri.Port)) {
                Log.WriteLog("UiO.TRAMP.MSRP.ConnectMSRPSync(): TCP connection with the remote host established.");
            } else {
                Log.WriteLog("UiO.TRAMP.MSRP.ConnectMSRPSync(): An error occured while trying to establish TCP connection with remote host. Aborting.");
                Abort();
            }
        }

        public void BindAndListenMSRPSync()
        {
            IPAddress ipAddress = null;

            Log.WriteLog("UiO.TRAMP.MSRP.BindAndListenMSRPSync(): Listening for the connection from the remote host.");

            try {
                ipAddress = IPAddress.Parse(localMSRPUri.DnsSafeHost);
            } catch (Exception exp) {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.BindAndListenMSRPSync(): An error occured while parsing remote IPAddress. Aborting.");
                Log.WriteLogWarning("UiO.TRAMP.MSRP.BindAndListenMSRPSync():" + exp.ToString());
                Abort();
            }

            int port;

            if (int.TryParse(c_default_msrp_port, out port)){
                Log.WriteLog("UiO.TRAMP.MSRP.BindAndListenMSRPSync(): Successfully parsed local port '" + port + "'.");
            } else {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.BindAndListenMSRPSync(): An error occured on transport protocol.");                
                Abort();
            }

            Log.WriteDebug("UiO.TRAMP.MSRP.BindAndListenMSRPSync(): Entering listening operation on the transport layer.");
            if (tcpConn.BindAndListenSync(ipAddress, port)) {
                Log.WriteLog("UiO.TRAMP.MSRP.BindAndListenMSRPSync(): TCP connection with the remote host established.");
            } else {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.BindAndListenMSRPSync(): An error occured on transport protocol.");                
                Abort();
            }
        }

        public void UpdateRemoteMSRPUri(String remoteUri) {
            remoteMSRPUri = new Uri(remoteUri);
        }        

        private String GetRandomString(int size) {
            Random rnd = new Random((int)DateTime.Now.Ticks);
            StringBuilder sb = new StringBuilder();

            char ch;
            for (int i = 0; i < size; i++) {
                ch = Convert.ToChar(Convert.ToInt32(Math.Floor(26 * rnd.NextDouble() + 65)));
                sb.Append(ch);
            }

            return sb.ToString();
        }

        public String GetRandomNumberString(int size) {
            string input = "0123456789";
            Random rnd = new Random((int)DateTime.Now.Ticks);

            StringBuilder sb = new StringBuilder();

            char ch;

            for (int i = 0; i < size; i++)
            {
                ch = input[rnd.Next(0, input.Length)];
                sb.Append(ch);
            }

            return sb.ToString();
        }

        private String getLine(String[] headerFields, String Line) {
            String wantedLine = "";
            Boolean found = false;

            foreach (String l in headerFields) {
                if (l.StartsWith(Line)) {
                    Log.WriteLog("UiO.TRAMP.MSRP.getLine(): Found header field '" + l + "' in message header.");
                    found = true;
                    wantedLine = l;
                }
            }

            if (!found) {
                Log.WriteLogWarning("UiO.TRAMP.MSRP.getLine(): Unable to find '" + Line + "' header field. Aborting.");
                Abort();
                return "";
            } else {
                return wantedLine;
            }
        }

        private String getToLine(string[] headerFields) {
            return getLine(headerFields, "To-Path");
        }

        private String getFromLine(string[] headerFields) {
            return getLine(headerFields, "From-Path");
        }

        private String getMessageIDLine(string[] headerFields) {
            return getLine(headerFields, "Message-ID");
        }

        private String getSuccessReportLine(string[] headerFields) {
            return getLine(headerFields, "Success-Report");
        }

        private String getFailureReportLine(string[] headerFields) {
            return getLine(headerFields, "Failure-Report");
        }
    }

    class MSRPSession {

        public Uri localUri;
        public Uri remoteUri;
        public String messageID;

        public Dictionary<string, string> transactionStates = new Dictionary<string, string>();

        public MSRPSession(Uri localUri, Uri remoteUri, String messageID) {
            this.localUri = localUri;
            this.remoteUri = remoteUri;
            this.messageID = messageID;
        }
    
    }

    class MSRPReportState {
        public int rStart;
        public int rEnd;
        public int tSize;

        public MSRPReportState(int rStart, int rEnd, int tSize) {
            this.rStart = rStart;
            this.rEnd = rEnd;
            this.tSize = tSize;
        
        }    
    }
    
}
