﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UiO.TRAMP.Logger;
using System.Net;
using UiO.TRAMP.Transport;
using UiO.TRAMP.SIP;
using UiO.TRAMP.State;
using System.Threading;

namespace UiO.TRAMP.RendezvousProtocol
{
    class RendezvousProtocolClass{

        private LoggerClass Log;
        public TCPConnection tcpConn;

        private SIPReceiver sipReceiver;
        private SIPSender sipSender;

        public SIPPayloadState payloadState;

        private Uri currentToURI;
        private Uri currentFromURI;

        

        public RendezvousProtocolClass (LoggerClass Log, TCPConnection tcpConn) {
            this.Log = Log;
            this.tcpConn = tcpConn;
            sipReceiver = new SIPReceiver(Log, tcpConn);
            sipSender = new SIPSender(Log, tcpConn);

            sipSender.mediaType = "application/sdp";
        }

        public void ReleaseResources() {
            Log.WriteDebug("UiO.TRAMP.RendezvousProtocol.ReleaseResources(): Releasing resoures on the rendezvous protocol layer.");
            if (tcpConn != null)
                tcpConn.ReleaseResources();
        }

        private void InitializePayload(byte[] payload, int payloadLength) {
            Log.WriteDebug("UiO.TRAMP.RendezvousProtocol.InitializePayload(): Initializing SIP payload with a payload of length '" + payloadLength + "'.");
            sipSender.payloadLength = payloadLength;
            sipSender.payload = payload;
        }       

        public void SendRendezvousRequest(IPAddress ipAddress, String selectedLocalIPAddress, byte[] payload, int payloadLength) {
            Log.WriteLog("UiO.TRAMP.RendezvousProtocol.SendRendezvousRequest(): Sending SIP INVITE to remote host.");

            currentToURI = new Uri("sip:" + ipAddress.ToString() + ";transport=tcp");
            currentFromURI = new Uri("sip:" + selectedLocalIPAddress + ";transport=tcp");

            // these logs might not belong here
            Log.WriteDebug("UiO.TRAMP.RendezvousProtocol.SendRendezvousRequest(): Setting the RequestURI/ToURI to: '" + currentToURI.ToString() + "'.");
            Log.WriteDebug("UiO.TRAMP.RendezvousProtocol.SendRendezvousRequest(): Setting the FromURI to: '" + currentFromURI.ToString() + "'.");

            InitializePayload(payload, payloadLength);  // SDP will come from a string so this function will need to be re-written to cover that case as well!
            sipSender.SIPSendINVITESync(currentToURI, currentFromURI);

            payloadState = sipSender.sipPayloadState;
        }

        public void ReceiveResposeAfterRendezvousRequest ()
        {
            Log.WriteDebug("UiO.TRAMP.RendezvousProtocol.ReceiveResposeAfterRendezvousRequest(): Receiving SIP response after SIP INVITE to the peer.");

            sipReceiver.sipPayloadState = payloadState;
            sipReceiver.SIPReceiveSync(true);

            Log.WriteLog("UiO.TRAMP.RendezvousProtocol.ReceiveResposeAfterRendezvousRequest(): Received SIP response from 'SIP transaction' layer. Passing state object.");
            payloadState = sipReceiver.sipPayloadState;
        }

        public void SendResponseToSessionCancelationRequest() {
            Log.WriteLog("UiO.TRAMP.RendezvousProtocol.SendResponseToSessionCancelationRequest(): Sending SIP 200 OK response to 'SIP transaction user' layer.");

            InitializePayload(null, 0);

            payloadState.content = null;
            payloadState.contentLength = 0;
            sipSender.sipPayloadState = payloadState;

            sipSender.SIPSend200Sync();
            payloadState = sipSender.sipPayloadState;  // this can happen to be overkill but still ...
        }

        public void ReceiveResponseAfterSessionCancellationRequest() {
            Log.WriteDebug("UiO.TRAMP.RendezvousProtocol.ReceiveResponseAfterSessionCancellationRequest(): Receiving SIP response after SIP BYE to the peer.");

            sipReceiver.sipPayloadState = payloadState;
            sipReceiver.SIPReceiveSync(true);

            Log.WriteLog("UiO.TRAMP.RendezvousProtocol.ReceiveResponseAfterSessionCancellationRequest(): Received SIP response from 'SIP transaction' layer. Passing state object.");
            payloadState = sipReceiver.sipPayloadState;
        }

        public void SendAckToRendezvousResponse () {
            Log.WriteLog("UiO.TRAMP.RendezvousProtocol.SendAckToRendezvousResponse(): Sending SIP ACK request. No content is present.");            
            InitializePayload(null, 0);
            sipSender.SIPSendACKSync();
        }

        public void SendSessionCancellationRequest() {
            Log.WriteLog("UiO.TRAMP.RendezvousProtocol.SendSessionCancellationRequest(): Sending SIP BYE request. No content is present.");
            InitializePayload(null, 0);
            sipSender.SIPSendBYESync(currentToURI, currentFromURI);
        }

        public void ReceiveRendezvousRequest() {
            Log.WriteDebug("UiO.TRAMP.RendezvousProtocol.ReceiveRendezvousRequest(): Receiving SIP INVITE request.");

            sipReceiver.SIPReceiveSync(true);

            Log.WriteLog("UiO.TRAMP.RendezvousProtocol.ReceiveRendezvousRequest(): Received SIP INVITE request from 'SIP transaction' layer. Passing state object.");
            payloadState = sipReceiver.sipPayloadState;    
        }

        public void ReceiveSessionCancelationRequest() {
            Log.WriteDebug("UiO.TRAMP.RendezvousProtocol.ReceiveSessionCancelationRequest(): Receiving SIP BYE request.");

            sipReceiver.SIPReceiveSync(true);

            Log.WriteDebug("UiO.TRAMP.RendezvousProtocol.ReceiveSessionCancelationRequest(): Received SIP BYE request from 'SIP transaction' layer. Passing state object.");
            payloadState = sipReceiver.sipPayloadState;
        }

        public void SendPositiveRendezvousResponse(byte [] payload, int payloadLength) {
            Log.WriteLog("UiO.TRAMP.RendezvousProtocol.SendPositiveRendezvousResponse(): Sending SIP 200 OK response on 'SIP transaction user' layer.");

            InitializePayload(payload, payloadLength);

            payloadState.content = null;
            payloadState.contentLength = 0;
            sipSender.sipPayloadState = payloadState;

            sipSender.SIPSend200Sync();
            payloadState = sipSender.sipPayloadState;  // this can happen to be overkill but still ...
        }

        public void SendNegativeRendezvousResponse(byte [] payload, int payloadLength) {
            Log.WriteLog("UiO.TRAMP.RendezvousProtocol.SendNegativeRendezvousResponse(): Sending SIP 400 Bad Request response on 'SIP transaction user' layer.");

            InitializePayload(payload, payloadLength);

            payloadState.content = null;
            payloadState.contentLength = 0;
            sipSender.sipPayloadState = payloadState;

            sipSender.SIPSend488Sync();
            payloadState = sipSender.sipPayloadState;  // this can happen to be overkill but still ...
        }

        public void ReceiveAckOnResponseToRendezvousRequest() {
            Log.WriteDebug("UiO.TRAMP.DataTransfer.ReceiveAckOnResponseToRendezvousRequest(): Receiving SIP ACK request as answer to SIP response to the peer.");

            sipReceiver.SIPReceiveSync(true);

            Log.WriteLog("UiO.TRAMP.DataTransfer.ReceiveAckOnResponseToRendezvousRequest(): Received SIP ACK request from 'SIP transaction' layer. Passing state object.");
            payloadState = sipReceiver.sipPayloadState;
        }
    }
}
