﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.Net.Sockets;
using System.Net;
using System.IO;
using UiO.TRAMP.Transport;
using UiO.TRAMP.Logger;
using UiO.TRAMP.State;
using UiO.TRAMP.SIP;

namespace UiO.TRAMP.NegotiationProtocol
{
    class NegotiationProtocolClass
    {
        private LoggerClass Log;        
        private TCPConnection tcpConn;
        public SIPPayloadState payloadState;

        private Uri currentToURI;
        private Uri currentFromURI;

        private SIPReceiver sipReceiver;
        private SIPSender sipSender;      

        public NegotiationProtocolClass(LoggerClass Log)
        {
            this.Log = Log;

            tcpConn = new TCPConnection(Log);
            sipReceiver = new SIPReceiver(Log, tcpConn);
            sipSender = new SIPSender(Log, tcpConn);

            sipSender.mediaType = "text/xml";
        }

        public TCPConnection getTCPConnection {
            get { return tcpConn; }
        }

        private void Abort()
        {
            ReleaseResources();
            Thread.CurrentThread.Abort();
        }

        public void ReleaseResources() {
            Log.WriteDebug("UiO.TRAMP.SIP.ReleaseResources(): Releasing resources on negotiation protocol layer.");
            tcpConn.ReleaseResources();
        }

        private void InitializePayload(byte [] payload, int payloadLength) {
            Log.WriteDebug("UiO.TRAMP.SIP.InitializePayload(): Initializing SIP payload with a payload of length '" + payloadLength + "'.");
            sipSender.payloadLength = payloadLength;
            sipSender.payload = payload;
        }

        public void BindAndListenNegotiationProtocolSync(IPAddress ipAddress, int localPort) {
            Log.WriteDebug("UiO.TRAMP.SIP.BindAndListenNegotiationProtocolSync(): Entering listening operation on 'SIP core' layer.");
            if (tcpConn.BindAndListenSync(ipAddress, localPort))
            {
                Log.WriteLog("UiO.TRAMP.SIP.BindAndListenNegotiationProtocolSync(): TCP connection with the remote host established.");
            }
            else {
                Log.WriteLogWarning("UiO.TRAMP.SIP.BindAndListenNegotiationProtocolSync(): An error occured on transport protocol.");                
                //tcpConn.Disconnect() - shutsdowns and closes socket that failed
                Abort();
            }
        }

        public void ConnectNegotiationProtocolSync(IPAddress ipAddress, int remotePort) {
            Log.WriteDebug("UiO.TRAMP.SIP.ConnectNegotiationProtocolSync(): Entering connect operation on 'SIP core' layer.");

            if (tcpConn.ConnectSync(ipAddress, remotePort)) {
                Log.WriteLog("UiO.TRAMP.SIP.ConnectNegotiationProtocolSync(): TCP connection with the remote host is established.");                
            } else {
                Log.WriteLogWarning("UiO.TRAMP.SIP.ConnectNegotiationProtocolSync(): An error occured on transport layer.");
                //tcpConn.Disconnect() - shutsdowns and closes socket that failed
                Abort();
            }
        }

        public void SendNegotiationRequest(IPAddress ipAddress, String selectedLocalIPAddress, byte[] payload, int payloadLength) {
            Log.WriteLog("UiO.TRAMP.SIP.SendNegotiationRequest(): Sending SIP INVITE to remote host.");

            currentToURI = new Uri("sip:" + ipAddress.ToString() + ";transport=tcp");
            currentFromURI = new Uri("sip:" + selectedLocalIPAddress + ";transport=tcp");

            // these logs might not belong here
            Log.WriteDebug("UiO.TRAMP.SIP.SendNegotiationRequest(): Setting the RequestURI/ToURI to: '" + currentToURI.ToString() + "'.");
            Log.WriteDebug("UiO.TRAMP.SIP.SendNegotiationRequest(): Setting the FromURI to: '" + currentFromURI.ToString() + "'.");

            InitializePayload(payload, payloadLength);  // SDP will come from a string so this function will need to be re-written to cover that case as well!
            sipSender.SIPSendINVITESync(currentToURI, currentFromURI);

            payloadState = sipSender.sipPayloadState;
        }

        public void SendAckToNegotiationResponse() {
            Log.WriteLog("UiO.TRAMP.SIP.SendAckToNegotiationResponse(): Sending SIP ACK request. No content is present.");
            sipSender.payloadLength = 0;
            sipSender.payload = null;
            InitializePayload(null, 0);
            sipSender.SIPSendACKSync();
        }

        public void SendPositiveNegotiationResponse(byte[] payload, int payloadLength) {
            Log.WriteLog("UiO.TRAMP.SIP.SendPositiveNegotiationResponse(): Sending SIP 200 OK response on 'SIP transaction user' layer.");

            InitializePayload(payload, payloadLength);

            payloadState.content = null;
            payloadState.contentLength = 0;
            sipSender.sipPayloadState = payloadState;

            sipSender.SIPSend200Sync();
            payloadState = sipSender.sipPayloadState;  // this can happen to be overkill but still ...
        }

        public void SendNegativeNegotiationResponse(byte[] payload, int payloadLength) {
            Log.WriteLog("UiO.TRAMP.SIP.SendNegativeNegotiationResponse(): Sending SIP 488 NOT ACCEPTABLE HERE response on 'SIP transaction user' layer.");

            InitializePayload(payload, payloadLength);

            payloadState.content = null;
            payloadState.contentLength = 0;
            sipSender.sipPayloadState = payloadState;

            sipSender.SIPSend488Sync();
            payloadState = sipSender.sipPayloadState;  // this can happen to be overkill but still ...
        }

        public void SendNegativeNegotiationResponseNotFound(byte[] payload, int payloadLength)
        {
            Log.WriteLog("UiO.TRAMP.SIP.SendNegativeNegotiationResponse(): Sending SIP 404 NOT FOUND response on 'SIP transaction user' layer.");

            InitializePayload(payload, payloadLength);

            payloadState.content = null;
            payloadState.contentLength = 0;
            sipSender.sipPayloadState = payloadState;

            sipSender.SIPSend404Sync();
            payloadState = sipSender.sipPayloadState;  // this can happen to be overkill but still ...
        }

        public void SendNegativeNegotiationResponseBadRequest(byte[] payload, int payloadLength)
        {
            Log.WriteLog("UiO.TRAMP.SIP.SendNegativeNegotiationResponseBadRequest(): Sending SIP 400 BAD REQUEST response on 'SIP transaction user' layer.");

            InitializePayload(payload, payloadLength);

            payloadState.content = null;
            payloadState.contentLength = 0;
            sipSender.sipPayloadState = payloadState;

            sipSender.SIPSend400Sync();
            payloadState = sipSender.sipPayloadState;  // this can happen to be overkill but still ...
        }

        public void ReceiveNegotiationRequest() {
            Log.WriteDebug("UiO.TRAMP.SIP.ReceiveNegotiationRequest(): Receiving SIP INVITE request.");

            sipReceiver.SIPReceiveSync(true);
            
            Log.WriteLog("UiO.TRAMP.SIP.ReceiveNegotiationRequest(): Received SIP INVITE request from 'SIP transaction' layer. Passing state object.");
            payloadState = sipReceiver.sipPayloadState;
        }

        public void ReceiveResposeAfterNegotiationRequest() {
            Log.WriteDebug("UiO.TRAMP.SIP.ReceiveResposeAfterNegotiationRequest(): Receiving SIP response after SIP INVITE to the peer.");

            sipReceiver.sipPayloadState = payloadState;
            sipReceiver.SIPReceiveSync(true);

            Log.WriteLog("UiO.TRAMP.SIP.ReceiveResposeAfterNegotiationRequest(): Received SIP response from 'SIP transaction' layer. Passing state object.");
            payloadState = sipReceiver.sipPayloadState;
        }

        public void ReceiveAckOnResponseToNegotiationRequest() {
            Log.WriteDebug("UiO.TRAMP.SIP.ReceiveAckOnResponseToNegotiationRequest(): Receiving SIP ACK request as answer to SIP response to the peer.");

            sipReceiver.SIPReceiveSync(true);

            Log.WriteLog("UiO.TRAMP.SIP.ReceiveAckOnResponseToNegotiationRequest(): Received SIP ACK request from 'SIP transaction' layer. Passing state object.");
            payloadState = sipReceiver.sipPayloadState;
        }
        
    }
}
