﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using UiO.TRAMP.Logger;
using UiO.TRAMP.Transport;
using System.IO;
using UiO.TRAMP.State;

namespace UiO.TRAMP.SIP
{
    class SIPReceiver
    {

        private LoggerClass Log;
        private TCPConnection tcpConn;      
        public SIPPayloadState sipPayloadState;

        private MemoryStream currentmStream;

        public SIPReceiver(LoggerClass Log, TCPConnection tcpConn) {
            this.Log = Log;
            this.tcpConn = tcpConn;
            this.currentmStream = new MemoryStream();

            sipPayloadState = new SIPPayloadState();
            sipPayloadState.dialog = new SIPDialog();
            sipPayloadState.transaction = new SIPTransaction();

        }

        private void Abort() {
            tcpConn.ReleaseResources();
            Thread.CurrentThread.Abort();
        }

        public void SIPReceiveSync(Boolean doPreparse)
        {

            int payloadFirstBytePosition = 0;                  
            /*
                Do not forget to start a timer that will fire after some time and close conenction and thread if not entire message has been received
             * When the message is received the timer will be disposed. The timer will acctually execute at the transport layer
             */
            
            do
            {
                MemoryStream mStream = new MemoryStream();                

                Log.WriteLog("UiO.TRAMP.SIP.SIPReceiveSync(): Calling 'Transport' layer to receive data.");
                if (tcpConn.ReceiveSync(mStream))
                {
                    Log.WriteDebug("UiO.TRAMP.SIP.SIPReceiveSync(): Transport layer returned '" + mStream.Length + "' B.");

                    try
                    {
                        byte[] buffer = new byte[mStream.Length];

                        mStream.Seek(0, SeekOrigin.Begin); //resetting memory stream position before read opeartion

                        int bytesFromStream = mStream.Read(buffer, 0, buffer.Length);

                        currentmStream.Write(buffer, 0, buffer.Length);

                        //dispose resources
                        buffer = null;
                        mStream.Dispose();

                        //Thread.CurrentThread.Abort();
                        if (doPreparse)
                        {
                            Log.WriteLog("UiO.TRAMP.SIP.SIPReceiveSync(): Starting SIPpreparser() function.");
                            payloadFirstBytePosition = SIPPreParser(currentmStream);
                            Log.WriteDebug("UiO.TRAMP.SIP.SIPReceiveSync(): Calculated firstByteposition is '" + payloadFirstBytePosition + "'.");
                        }
                        else
                        {
                            return; // when we only query for the payload data the outher loop will take care of multiple calls of SIPReceiveSyn() function
                        }
                    }
                    catch (Exception exp) {
                        Log.WriteLogWarning(exp.ToString());
                    }
                } else {
                    Log.WriteLogWarning("UiO.TRAMP.SIP.SIPReceiveSync(): An error occured on transport layer while receiving data. Aborting.");
                    //tcpConn.Disconnect() - shutsdowns and closes socket that failed                    
                    Abort();
                }

            } while (payloadFirstBytePosition < 0);

            /*
                From this point on we have two thigs:
             * a) enough of the input stream to have the complete header of the SIP message
             * b) position of the first byte of the payload in the message
             */

            // here comes the general SIP parsing
            SIPParse(currentmStream, payloadFirstBytePosition);
        }
        /// <summary>
        /// Used priamrly to find the position in the received stream where the payload begins. It will ask for more data from the socket if 
        /// not enough bytes are received in the first round to complete the SIP header
        /// </summary>
        /// <param name="mStream"></param>
        private int SIPPreParser(MemoryStream currentmStream)
        {

            int position = 0; // position in stream
            int streamsCurrentPosition = 0;


            if (currentmStream.Length == 0) {
                Log.WriteLogWarning("UiO.TRAMP.SIP.SIPPreParser(): Found no data in the receive buffer. Aborting.");
                //tcpConn.Disconnect() - shutsdowns and closes socket that failed                
                Abort();
            } else {
                streamsCurrentPosition = (int)currentmStream.Position;
                Log.WriteDebug("UiO.TRAMP.SIP.SIPPreparser(): Current stream position is '" + streamsCurrentPosition + "'.");
            }

            currentmStream.Seek(0, SeekOrigin.Begin); //resetting current streams' position so that it could be analysed

            while (HasMoreData(currentmStream))
            {
                char ch = Convert.ToChar(currentmStream.ReadByte());
                
                switch (position)
                {
                    case 0:                        
                        if (ch == '\r') { position++;}
                        break;
                    case 1:
                        if (ch == '\n') { position++; }
                        else {
                            ResetData(currentmStream, ref position);
                        }
                        break;
                    case 2:
                        if (ch == '\r') { position++; }
                        else {
                            ResetData(currentmStream, ref position);
                        }
                        break;
                    case 3:
                        if (ch == '\n') {
                            // currentStream.Position holds the first byte of the payload, if any
                            position = (int)currentmStream.Position;

                            // the stream points where it did before the function started. This is neccasary in case SIPParse function must receive more data from transport layer
                            currentmStream.Position = (long)streamsCurrentPosition; 

                            Log.WriteDebug("UiO.TRAMP.SIP.SIPPreparser(): Found double 'new line' in the input stream. The payload begins at position '" + position + "' in the stream.");

                            return position;
                        } else {
                            ResetData(currentmStream, ref position);
                        }
                        break;
                    default:
                        break;
                }
            }

            /*
             * If we came so far the two only possibilities are that not enough data has been read to encompas entire header
             * or that the request is not formed correctly (meaning no double new-line is found in the received string). 
             * The latter wiill not be considered, at least not for now, but the former can be remedied with extra read from transport layer
             */
            currentmStream.Position = (long)streamsCurrentPosition;
            return -1;
        }

        private Boolean HasMoreData(MemoryStream mStream)
        {
            return (int)(mStream.Length - mStream.Position) > 0;
        }

        private void ResetData(MemoryStream mStream, ref int postition)
        {
            postition = 0;

            int currentPosition = (int)mStream.Position;
            if (currentPosition != 0) { mStream.Position = (long)(currentPosition - 1); }
        }

        private void SIPParse(MemoryStream currentmStream, int payloadPosition)
        {
            int lastHeaderByte = payloadPosition - 3; //-3 to point to the last LF character before CRLF that separates header and payload
            int currentStreamsPosition = (int)currentmStream.Position;

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse(): extracting SIP header from the input stream. The last header byte is found at position '" + lastHeaderByte  + "'.");
            currentmStream.Seek(0, SeekOrigin.Begin);

            byte[] headerBytes = new byte[lastHeaderByte];
            StringBuilder sb = new StringBuilder();

            int bytesRead = currentmStream.Read(headerBytes, 0, headerBytes.Length);
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse(): Read '" + bytesRead + "' B from memory stream and into the StringBuilder.");

            //sb now holds all the received data as UTF8 string
            sb.Append(new UTF8Encoding().GetString(headerBytes));

            String[] headerFields = sb.ToString().Split(new string[] { Environment.NewLine }, StringSplitOptions.None);

            //puting back stream to its current position in case more reads are needed from the transport layer
            currentmStream.Position = currentStreamsPosition;

            if (headerFields[0].StartsWith("SIP"))
            {
                //Response is received
                Log.WriteLogInformational("UiO.TRAMP.SIP.SIPParse(): Received SIP response.");
                SIPParseResponse(currentmStream, payloadPosition, headerFields);
            } else { 
                //Request is received
                Log.WriteLogInformational("UiO.TRAMP.SIP.SIPParse(): Received SIP request.");
                SIPParseRequest(currentmStream, payloadPosition, headerFields);
            }

            // we need to reset the currentStream so that it is ready when it receives new data in the next round
            currentmStream.SetLength(0);
        }

        private void SIPSendSync()
        {
            throw new NotImplementedException();
        }


        private void SIPParseResponse(MemoryStream currentmStream, int payloadPosition, string [] headerFields) {
            int contentLength = 0;
            Boolean hFound = false;
            //first we need to determine if the entire payload has been received (by now we know that at least entire header is received)
            foreach (String header in headerFields)
            {
                if (header.StartsWith("Content-Length"))
                {
                    Log.WriteDebug("UiO.TRAMP.SIP.SIPParseResponse(): Found Content-Length header field with value of '" + contentLength + "' B.");
                    hFound = true;
                    if (int.TryParse(header.Split(':')[1].Trim(), out contentLength))
                    {
                        Log.WriteDebug("UiO.TRAMP.SIP.SIPParseResponse(): Extracted content length form header field is '" + contentLength + "'.");
                    }
                    else
                    {
                        Log.WriteLogWarning("UiO.TRAMP.SIP.SIPParseResponse(): Unable to find 'Content-Length' header field in the input string. Aborting.");
                        //Here we will be forced to send a response of type 4xx to signal the other side that something went wrong. For now it is just an abort.                        
                        Abort();
                    }
                }
            }
            if (!hFound)
            {
                Log.WriteLogWarning("UiO.TRAMP.SIP.SIPParseResponse(): Unable to find 'Content-Length' header field in input stream. Aborting.");                
                Abort();
            }

            
            if (contentLength > 0) {
                if (currentmStream.Length < (payloadPosition + contentLength)) {
                    while ((currentmStream.Length - payloadPosition) < contentLength) {
                        // in the wors case scenarion this potentially infinite loop should be interrupted by a transaction timer                   
                        Log.WriteDebug("UiO.TRAMP.SIP.SIPParseResponse(): Calculated length is: " + (currentmStream.Length - payloadPosition));
                        Log.WriteDebug("UiO.TRAMP.SIP.SIPParseResponse(): Length from Content-Length header field is: " + contentLength);
                        Log.WriteDebug("UiO.TRAMP.SIP.SIPParseResponse(): Not all payload data has been received. Returning to socket for more data.");
                        SIPReceiveSync(false);
                    }
                } else {
                    Log.WriteDebug("UiO.TRAMP.SIP.SIPParseResponse(): current stream contains entire message. Exiting receive loop.");
                }
            } else {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParseResponse(): Content-Length is 0, no need to check for more data on the connection.");
            }
            
            if (headerFields[0].StartsWith("SIP/2.0 200")) {
                SIPParse200Response(currentmStream, payloadPosition, headerFields, contentLength);
            } else if (headerFields[0].StartsWith("SIP/2.0 400")) {
                SIPParse400Response(currentmStream, payloadPosition, headerFields, contentLength);
            } else if (headerFields[0].StartsWith("SIP/2.0 404")) {
                SIPParse404Response(currentmStream, payloadPosition, headerFields, contentLength);
            } else if (headerFields[0].StartsWith("SIP/2.0 488")) {
                SIPParse488Response(currentmStream, payloadPosition, headerFields, contentLength);
            } else {
                Log.WriteLogWarning("UiO.TRAMP.SIP.SIPParseResponse(): Received an unrecognized method '" + headerFields[0].Split(' ')[0] + "'. Aborting.");
                // here as well the most appropriate would be to return a 4xx class response and finish transaction. For now just abort                
                Abort();
            }

            writeStream(currentmStream);
        }

        private void SIPParseRequest(MemoryStream currentmStream, int payloadPosition, string [] headerFields){
            
            int contentLength = 0;
            Boolean hFound = false;
            
            //first we need to determine if the entire payload has been received (by now we know that at least entire header is received)
            foreach(String header in headerFields){
                if (header.StartsWith("Content-Length")){
                    
                    hFound = true;
                    
                    if (int.TryParse(header.Split(':')[1].Trim(), out contentLength))
                    {
                        Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequest(): Found Content-Lengt header field with value of '" + contentLength + "' B.");
                    } else {
                        Log.WriteLogWarning("UiO.TRAMP.SIP.SIPParseRequest(): Unable to extract value from Content-Length header field. Aborting.");
                        //Here we will be forced to send a response of type 4xx to signal the other side that something went wrong. For now it is just an abort.                        
                        Abort();
                    }
                }
            }
            if (!hFound) {
                Log.WriteLogWarning("UiO.TRAMP.SIP.SIPParseRequest(): Unable to find 'Content-Length' header field in the input string. Aborting.");                
                Abort();
            }

            //receive loop
            
            if (contentLength > 0) {
                if (currentmStream.Length < (payloadPosition + contentLength)) {
                    while ((currentmStream.Length - payloadPosition) < contentLength) {
                        // in the worst case scenarion this potentially infinite loop should be interrupted by a transaction timer                   
                        Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequest(): Calculated length is: '" + (currentmStream.Length - payloadPosition) + "' B.");
                        Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequest(): Length from Content-Length header field is: " + contentLength + "' B.");
                        Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequest(): Not all payload data has been received. Returning to socket for more data.");
                        SIPReceiveSync(false);
                    }
                } else {
                    Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequest(): Current stream contains entire message. No need to check for more data on the transport layer.");
                }
            } else {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequest(): Received request without the payload (Content-Length = 0). No need to check for more data on the transport layer.");
            }
            
            if (headerFields[0].StartsWith("INVITE")) {
                SIPParseRequestInvite(currentmStream, payloadPosition, headerFields, contentLength);
            }
            else if (headerFields[0].StartsWith("ACK")) {
                SIPParseRequestAck(currentmStream, payloadPosition, headerFields);
            }
            else if (headerFields[0].StartsWith("BYE"))
            {
                SIPParseRequestBye(currentmStream, payloadPosition, headerFields);
            }
            else {
                Log.WriteLogWarning("UiO.TRAMP.SIP.SIPParseRequest(): Received an unrecognized method '" + headerFields[0].Split(' ')[0] + "'. Aborting.");
                // here as well the most appropriate would be to return a 4xx class response and finish transaction. For now just abort                
                Abort();
            }


            writeStream(currentmStream);
        }

        private void SIPParse488Response(MemoryStream curentStream, int payloadPosition, string[] headerFields, int contentLength) {
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse488Response(): Received '488 Not Acceptable Here' response.");

            // See if the received response matches the current transaction
            String transactionID = getViaLine(headerFields).Split(';')[1].Split('=')[1];

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse488Response(): Extracted '" + transactionID + "' from the '488 Not Acceptable Here' response.");

            if (transactionID.Equals(sipPayloadState.transaction.transactionID, StringComparison.Ordinal)) {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse488Response(): Received transactionID matches the current transactionID: '" + transactionID + " == " + sipPayloadState.transaction.transactionID + "'.");                
                sipPayloadState.rType = SIPPayloadState.receivedResponseType.NOTACCEPTABLEHERE; // mark the type of received response
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse488Response(): Marked the received response as '488 Not Acceptable Here'.");
            } else {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse488Response(): Received transactionID does not match the current transactionID: '" + transactionID + " != " + sipPayloadState.transaction.transactionID + "'.");                
                Abort();
            }

            String callID = getCallIDLine(headerFields).Split(':')[1].Trim();
            String localTag = getFromLine(headerFields).Split(';')[1].Split('=')[1];
            String remoteTag = getToLine(headerFields).Split(';')[1].Split('=')[1];

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse488Response(): Retrieved Call-ID from '488 Not Acceptable Here' is: '" + callID + "'. Exiting Call-ID is: '" + sipPayloadState.dialog.callID + "'.");
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse488Response(): Retrieved Local-Tag from '488 Not Acceptable Here' is: '" + localTag + "'. Existing local-Tag is: '" + sipPayloadState.dialog.localTag + "'.");
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse488Response(): Retrieved Remote-Tag from '488 Not Acceptable Here' is: '" + remoteTag + "'. Existing remote-Tag is: '" + sipPayloadState.dialog.remoteTag + "'.");

            if ((callID.Equals(sipPayloadState.dialog.callID, StringComparison.Ordinal)) && (localTag.Equals(sipPayloadState.dialog.localTag, StringComparison.Ordinal))) {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse488Response(): Received response matches the current pre-dialog.");
                // Completing the dialog 
                sipPayloadState.dialog.remoteTag = remoteTag;
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse488Response(): Completing the existing dialog with remote-Tag '" + remoteTag + "'.");
            } else {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse488Response(): Received request does not match the current pre-dialog. Aborting.");                
                Abort();
            }

            byte[] payload = new byte[contentLength];
            currentmStream.Seek(payloadPosition, SeekOrigin.Begin);
            currentmStream.Read(payload, 0, contentLength);

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse488Response(): ConntentLength: " + contentLength);

            sipPayloadState.content = payload;
            sipPayloadState.contentLength = contentLength;
        }

        private void SIPParse400Response(MemoryStream curentStream, int payloadPosition, string[] headerFields, int contentLength)
        {
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse400Response(): Received '400 Bad Request' response.");

            // See if the received response matches the current transaction
            String transactionID = getViaLine(headerFields).Split(';')[1].Split('=')[1];

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse400Response(): Extracted '" + transactionID + "' from the '400 Bad Request' response.");

            if (transactionID.Equals(sipPayloadState.transaction.transactionID, StringComparison.Ordinal))
            {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse400Response(): Received transactionID matches the current transactionID: '" + transactionID + " == " + sipPayloadState.transaction.transactionID + "'.");
                sipPayloadState.rType = SIPPayloadState.receivedResponseType.BADREQUEST; // mark the type of received response
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse400Response(): Marked the received response as '400 Bad Request'.");
            }
            else
            {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse400Response(): Received transactionID does not match the current transactionID: '" + transactionID + " != " + sipPayloadState.transaction.transactionID + "'.");                
                Abort();
            }

            String callID = getCallIDLine(headerFields).Split(':')[1].Trim();
            String localTag = getFromLine(headerFields).Split(';')[1].Split('=')[1];
            String remoteTag = getToLine(headerFields).Split(';')[1].Split('=')[1];

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse400Response(): Retrieved Call-ID from '400 Bad Request' is: '" + callID + "'. Exiting Call-ID is: '" + sipPayloadState.dialog.callID + "'.");
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse400Response(): Retrieved Local-Tag from '400 Bad Request' is: '" + localTag + "'. Existing local-Tag is: '" + sipPayloadState.dialog.localTag + "'.");
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse400Response(): Retrieved Remote-Tag from '400 Bad Request' is: '" + remoteTag + "'. Existing remote-Tag is: '" + sipPayloadState.dialog.remoteTag + "'.");

            if ((callID.Equals(sipPayloadState.dialog.callID, StringComparison.Ordinal)) && (localTag.Equals(sipPayloadState.dialog.localTag, StringComparison.Ordinal)))
            {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse400Response(): Received response matches the current pre-dialog.");
                // Completing the dialog 
                sipPayloadState.dialog.remoteTag = remoteTag;
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse400Response(): Completing the existing dialog with remote-Tag '" + remoteTag + "'.");
            }
            else
            {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse400Response(): Received request does not match the current pre-dialog. Aborting.");                
                Abort();
            }

            byte[] payload = new byte[contentLength];
            currentmStream.Seek(payloadPosition, SeekOrigin.Begin);
            currentmStream.Read(payload, 0, contentLength);

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse400Response(): ConntentLength: " + contentLength);

            sipPayloadState.content = payload;
            sipPayloadState.contentLength = contentLength;
        }

        private void SIPParse404Response(MemoryStream curentStream, int payloadPosition, string[] headerFields, int contentLength)
        {
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse404Response(): Received '404 Not Found' response.");

            // See if the received response matches the current transaction
            String transactionID = getViaLine(headerFields).Split(';')[1].Split('=')[1];

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse404Response(): Extracted '" + transactionID + "' from the '404 Not Found' response.");

            if (transactionID.Equals(sipPayloadState.transaction.transactionID, StringComparison.Ordinal))
            {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse404Response(): Received transactionID matches the current transactionID: '" + transactionID + " == " + sipPayloadState.transaction.transactionID + "'.");
                sipPayloadState.rType = SIPPayloadState.receivedResponseType.BADREQUEST; // mark the type of received response
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse404Response(): Marked the received response as '404 Not Found'.");
            }
            else
            {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse404Response(): Received transactionID does not match the current transactionID: '" + transactionID + " != " + sipPayloadState.transaction.transactionID + "'.");                
                Abort();
            }

            String callID = getCallIDLine(headerFields).Split(':')[1].Trim();
            String localTag = getFromLine(headerFields).Split(';')[1].Split('=')[1];
            String remoteTag = getToLine(headerFields).Split(';')[1].Split('=')[1];

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse404Response(): Retrieved Call-ID from '404 Not Found' is: '" + callID + "'. Exiting Call-ID is: '" + sipPayloadState.dialog.callID + "'.");
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse404Response(): Retrieved Local-Tag from '404 Not Found' is: '" + localTag + "'. Existing local-Tag is: '" + sipPayloadState.dialog.localTag + "'.");
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse404Response(): Retrieved Remote-Tag from '404 Not Found' is: '" + remoteTag + "'. Existing remote-Tag is: '" + sipPayloadState.dialog.remoteTag + "'.");

            if ((callID.Equals(sipPayloadState.dialog.callID, StringComparison.Ordinal)) && (localTag.Equals(sipPayloadState.dialog.localTag, StringComparison.Ordinal))) {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse404Response(): Received response matches the current pre-dialog.");
                // Completing the dialog 
                sipPayloadState.dialog.remoteTag = remoteTag;
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse404Response(): Completing the existing dialog with remote-Tag '" + remoteTag + "'.");
            } else {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse404Response(): Received request does not match the current pre-dialog. Aborting.");                
                Abort();
            }

            byte[] payload = new byte[contentLength];
            currentmStream.Seek(payloadPosition, SeekOrigin.Begin);
            currentmStream.Read(payload, 0, contentLength);

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse404Response(): ConntentLength: " + contentLength);

            sipPayloadState.content = payload;
            sipPayloadState.contentLength = contentLength;
        }

        private void SIPParse200Response(MemoryStream curentStream, int payloadPosition, string[] headerFields, int contentLength)
        {
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse200Response(): Received '200 OK' response.");

            // See if the received response matches the current transaction
            String transactionID = getViaLine(headerFields).Split(';')[1].Split('=')[1];

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse200Response(): Extracted '" + transactionID + "' from the '200 OK' response.");

            if (transactionID.Equals(sipPayloadState.transaction.transactionID, StringComparison.Ordinal)) {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse200Response(): Received transactionID matches the current transactionID: '" + transactionID + " == " + sipPayloadState.transaction.transactionID + "'.");
                sipPayloadState.transaction = null; //RFC 3261 says that 200 OK response imidiatelly destroyes ongoing transation and that following ack is sent beyond any transaction (directly form TU, but branch parameter is still there)
                sipPayloadState.rType = SIPPayloadState.receivedResponseType.OK; // mark the type of received response
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse200Response(): Marked the received response as '200 OK'.");
            } else {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse200Response(): Received transactionID does not match the current transactionID: '" + transactionID + " != " + sipPayloadState.transaction.transactionID + "'.");                
                Abort();
            }

            String callID = getCallIDLine(headerFields).Split(':')[1].Trim();
            String localTag = getFromLine(headerFields).Split(';')[1].Split('=')[1];
            String remoteTag = getToLine(headerFields).Split(';')[1].Split('=')[1];

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse200Response(): Retrieved Call-ID from '200 OK' is: '" + callID + "'. Exiting Call-ID is: '" + sipPayloadState.dialog.callID + "'.");
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse200Response(): Retrieved Local-Tag from '200 OK' is: '" + localTag + "'. Existing local-Tag is: '" + sipPayloadState.dialog.localTag + "'.");
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse200Response(): Retrieved Remote-Tag from '200 OK' is: '" + remoteTag + "'. Existing remote-Tag is: '" + sipPayloadState.dialog.remoteTag + "'.");

            if ((callID.Equals(sipPayloadState.dialog.callID, StringComparison.Ordinal)) && (localTag.Equals(sipPayloadState.dialog.localTag, StringComparison.Ordinal))) {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse200Response(): Received response matches the current pre-dialog.");
                // Completing the dialog 
                sipPayloadState.dialog.remoteTag = remoteTag;
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse200Response(): Completing the existing dialog with remote-Tag '" + remoteTag + "'.");
            } else {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParse200Response(): Received request does not match the current pre-dialog. Aborting.");                
                Abort();
            }

            byte[] payload = new byte[contentLength];
            currentmStream.Seek(payloadPosition, SeekOrigin.Begin);
            currentmStream.Read(payload, 0, contentLength);

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParse200Response(): ConntentLength: " + contentLength);

            sipPayloadState.content = payload;
            sipPayloadState.contentLength = contentLength;
        }

        private void SIPParseRequestAck(MemoryStream currentmStream, int payloadPosition, String[] headerFields)
        {
            Log.WriteLogInformational("UiO.TRAMP.SIP.SIPParseRequestAck(): Received ACK request.");

            try
            {
                String callID = getCallIDLine(headerFields).Split(':')[1].Trim();
                String localTag = getToLine(headerFields).Split(';')[1].Split('=')[1];
                String remoteTag = getFromLine(headerFields).Split(';')[1].Split('=')[1];

                Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestAck(): Retrieved Call-ID from ACK: '" + callID + "'. Exiting Call-ID is: '" + sipPayloadState.dialog.callID + "'.");
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestAck(): Retrieved local-Tag from ACK: '" + localTag + "'. Existing local-Tag is: '" + sipPayloadState.dialog.localTag + "'.");
                Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestAck(): Retrieved remote-Tag from ACK: '" + remoteTag + "'. Existing remote-Tag is: '" + sipPayloadState.dialog.remoteTag + "'.");

                if ((callID.Equals(sipPayloadState.dialog.callID, StringComparison.Ordinal)) && (localTag.Equals(sipPayloadState.dialog.localTag, StringComparison.Ordinal)) && (remoteTag.Equals(sipPayloadState.dialog.remoteTag, StringComparison.Ordinal))) {
                    
                    Log.WriteLogInformational("UiO.TRAMP.SIP.SIPParseRequestAck(): Received ACK request matches the current dialog.");
                    if (sipPayloadState.transaction != null) {
                        Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestAck(): Ack is response to a non-200 OK SIP response from the peer. Checking for transactionID match.");

                        String branch = getViaLine(headerFields).Split(';')[1].Split('=')[1].Trim();
                        Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestAck(): Retrieved branch parameter from ACK: '" + branch + "'. Existing branch parameter is: '" + sipPayloadState.transaction.transactionID + "'.");
                        if (branch.Equals(sipPayloadState.transaction.transactionID, StringComparison.Ordinal))
                        {
                            Log.WriteLogInformational("UiO.TRAMP.SIP.SIPParseRequestAck(): Received ACK Request matchs a current transaction with transactionID ='" + sipPayloadState.transaction.transactionID + "'.");
                            // when the ACK is received the current transaction ends anyway
                            sipPayloadState.transaction = null;
                        } else {
                            Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestAck(): Received ACK request to a non-200 response does not match the current transaction. Aborting.");                            
                            Abort();
                        }
                    } else {
                        Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestAck(): Ack is response to a non-200 OK SIP response from the peer. It represents a transacion in its own.");
                    }
                } else {
                    Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestAck(): Received ACK request does not match the current pre-dialog. Aborting.");                    
                    Abort();
                }
            }
            catch (Exception exp) {
                Log.WriteLogWarning(exp.ToString());
            }            
        }

        private void SIPParseRequestInvite (MemoryStream currentmStream, int payloadPosition, String [] headerFields, int contentLength){

            /* Here we need to do the following:
             * 1) Create transaction
             * 2) Create the dialog (or at least a half of it)
             * 3) extract all of this information to the "SIP Transaction User" and "negotiation layer" for further processing
             */

            Log.WriteLogInformational("UiO.TRAMP.SIP.SIPParseRequestInvite(): Received INVITE request.");

            String viaLine = getViaLine(headerFields);
            
            sipPayloadState.transaction.transactionID = viaLine.Split(';')[1].Split('=')[1];
            sipPayloadState.transaction.viaLine = viaLine;

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestInvite(): 'Via' header field: '" + viaLine + "'."); 
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestInvite(): Extracted transactionID: '" + sipPayloadState.transaction.transactionID + "'.");

            String fromLine = getFromLine(headerFields);
            String toLine = getToLine(headerFields);
            String CSeqLine = getCSeqLine(headerFields);
            String contactLine = getContactLine(headerFields);
            String callIDLine = getCallIDLine(headerFields);

            sipPayloadState.dialog.callID = callIDLine.Split(':')[1].Trim();
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestInvite(): Call-ID: '" + sipPayloadState.dialog.callID + "'.");

            sipPayloadState.dialog.remoteTag = fromLine.Split(';')[1].Split('=')[1];
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestInvite(): Remote-Tag: '" + sipPayloadState.dialog.remoteTag + "'.");
            sipPayloadState.dialog.localTag = "";  // at this point we do not have a local tag
            
            sipPayloadState.dialog.localUri = toLine.Split(' ')[1].Trim();            
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestInvite(): Local-Uri: '" + sipPayloadState.dialog.localUri + "'.");

            sipPayloadState.dialog.remoteUri = fromLine.Split(';')[0].Split(' ')[1].Trim();  // not exactly perfect
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestInvite(): Remote-Uri: '" + sipPayloadState.dialog.remoteUri + "'.");

            sipPayloadState.dialog.remoteSeqNumber = CSeqLine.Split(':')[1].Trim().Split(' ')[0];
            sipPayloadState.dialog.localSeqNumber = ""; // since we have not sent a request from this host yet, we do not have local seq number
            sipPayloadState.transaction.method = CSeqLine.Split(':')[1].Trim().Split(' ')[1];
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestInvite(): Remote-Seq-Number: '" + sipPayloadState.dialog.remoteSeqNumber + "'.");

            sipPayloadState.dialog.remoteTarget = contactLine.Split(' ')[1].Substring(1, contactLine.Split(' ')[1].Length - 2);  //repair
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestInvite(): Remote-Target '" + sipPayloadState.dialog.remoteTarget + "'.");

            byte[] payload = new byte[contentLength];
            currentmStream.Seek(payloadPosition, SeekOrigin.Begin);
            currentmStream.Read(payload, 0, contentLength);

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestInvite(): Entire payload length is: '" + contentLength + "' B.");

            //writeStream(currentmStream);

            sipPayloadState.content = payload;
            sipPayloadState.contentLength = contentLength;

        }

        private string getContactLine(String[] headerFields)
        {
            return getLine(headerFields, "Contact");
        }

        private string getToLine(String[] headerFields)
        {
            return getLine(headerFields, "To");
        }

        private string getCSeqLine(String[] headerFields)
        {
            return getLine(headerFields, "CSeq");
        }

        private string getCallIDLine(String[] headerFields)
        {
            return getLine(headerFields, "Call-ID");
        }

        private string getFromLine(String[] headerFields)
        {
            return getLine(headerFields, "From");
        }

        private string getViaLine(String[] headerFields) {
            return getLine(headerFields, "Via");
        }

        private String getLine(String[] headerFields, String Line) {
            String wantedLine = "";
            Boolean found = false;

            foreach (String l in headerFields) {
                if (l.StartsWith(Line))
                {
                    Log.WriteLog("UiO.TRAMP.SIP.getLine(): Found header field '" + l + "' in message header.");
                    found = true;
                    wantedLine = l;
                }
            }

            if(!found){
                Log.WriteLogWarning("UiO.TRAMP.SIP.getLine(): Unable to find '" + Line + "' header field. Aborting.");                
                Abort();
                return "";
            } else {
                return wantedLine;
            }
        }

        private void SIPParseRequestBye(MemoryStream currentmStream, int payloadPosition, String[] headerFields)
        {
            /* Here we need to do the following:
             * 1) Create transaction            
             * 2) extract all of this information to the "SIP Transaction User" and "negotiation layer" for further processing
             */

            Log.WriteLogInformational("UiO.TRAMP.SIP.SIPParseRequestBye(): Received BYE request.");

            String viaLine = getViaLine(headerFields);

            sipPayloadState.transaction = new SIPTransaction();

            sipPayloadState.transaction.transactionID = viaLine.Split(';')[1].Split('=')[1];
            sipPayloadState.transaction.viaLine = viaLine;

            Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestBye(): 'Via' header field: '" + viaLine + "'.");
            Log.WriteDebug("UiO.TRAMP.SIP.SIPParseRequestBye(): Extracted transactionID: '" + sipPayloadState.transaction.transactionID + "'.");

            String fromLine = getFromLine(headerFields);
            String toLine = getToLine(headerFields);
            String CSeqLine = getCSeqLine(headerFields);
            //String contactLine = getContactLine(headerFields); - BYE request does not contain and does not need contact line
            String callIDLine = getCallIDLine(headerFields);

            sipPayloadState.transaction.method = CSeqLine.Split(':')[1].Trim().Split(' ')[1];                                    

            //writeStream(currentmStream);

            sipPayloadState.content = null;
            sipPayloadState.contentLength = 0;
        }

        private void writeStream(MemoryStream mStream) {
            byte[] payload = new byte[mStream.Length];
            int position = (int)mStream.Position;
            mStream.Seek(0, SeekOrigin.Begin);
            mStream.Read(payload, 0, payload.Length);
            mStream.Position = (long)position;
            StringBuilder sb = new StringBuilder();
            sb.Append(new UTF8Encoding().GetString(payload));
            Log.WriteLog("SIPParse(): Received the following content: \r\n\r\n" + sb.ToString());
        }
            

    }
}
