﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UiO.TRAMP.Logger;
using UiO.TRAMP.Transport;
using System.IO;
using System.Threading;
using UiO.TRAMP.State;

namespace UiO.TRAMP.SIP
{
    class SIPSender
    {

        private LoggerClass Log;
        private TCPConnection tcpConn;

        public SIPPayloadState sipPayloadState;

        private byte[] _payloadByteArray;
        private int _payloadLength;

        public String mediaType;

        public byte [] payload
        {
            get
            {
                return _payloadByteArray;
            }

            set
            {
                _payloadByteArray = value;
            }
        }

        public int payloadLength
        {
            get
            {
                return _payloadLength;
            }

            set
            {
                _payloadLength = value;
            }
        }

        public SIPSender(LoggerClass Log, TCPConnection tcpConn) {
            this.Log = Log;
            this.tcpConn = tcpConn;

            sipPayloadState = new SIPPayloadState ();
            sipPayloadState.dialog = new SIPDialog ();
            sipPayloadState.transaction = new SIPTransaction ();
        }

        private void Abort() {
            tcpConn.ReleaseResources();
            Thread.CurrentThread.Abort();
        }

        public void SIPSendINVITESync(Uri currentToUri, Uri currentFromUri) {            
            SIPSendRequestSync(SIPRequestType.INVITE, currentToUri, currentFromUri);
        }

        public void SIPSendACKSync()
        {
            SIPSendRequestSync(SIPRequestType.ACK, null, null);
        }

        public void SIPSendBYESync(Uri currentToUri, Uri currentFromUri)
        {
            SIPSendRequestSync(SIPRequestType.BYE, currentToUri, currentFromUri);
        }

        private void SIPSendRequestSync(SIPRequestType method, Uri currentToUri, Uri currentFromUri) {

            StringBuilder header = new StringBuilder();
            Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): entering function for method '" + Enum.GetName(typeof(SIPRequestType), method) + "' and creating header.");

            switch (method)
            {
                case SIPRequestType.INVITE:
                    //Request line
                    AddRequestLine(SIPRequestType.INVITE, header, currentToUri);
                    //Via line
                    AddViaHeader(header, currentFromUri, true);
                    //Max-Forwards line
                    AddMaxForwards(header);
                    //To line
                    AddToHeader(header, currentToUri);
                    sipPayloadState.dialog.remoteTag = "";
                    //From line
                    AddFromHeader(header, currentFromUri);
                    //Call-ID
                    AddCallIDHeader(header, true);
                    //CSeq line
                    AddCSeqHeader(SIPRequestType.INVITE, header, true);
                    //Contact line
                    AddContactHeader(header, currentFromUri);
                    //Content-Type line
                    AddContentTypeHeader(header, mediaType);  // for now, this will be variable as SIP will be used as randezvouis protocol for MSRP as well
                    //Content-Length line
                    AddContentLength(header, _payloadLength);
                    header.Append("\r\n");                    
                    break;
                case SIPRequestType.ACK:
                    //Request Line
                    header.Append(Enum.GetName(typeof(SIPRequestType), method) + " " + sipPayloadState.dialog.remoteUri + " " + "SIP/2.0");
                    header.Append("\r\n");
                    //via-line
                    if (sipPayloadState.transaction == null) {
                        AddViaHeader(header, new Uri(sipPayloadState.dialog.localUri), true);
                    } else {
                        AddViaHeader(header, new Uri(sipPayloadState.dialog.localUri), false);
                    }
                    //max-forwards line
                    AddMaxForwards(header);
                    //To-line
                    header.Append("To: " + sipPayloadState.dialog.remoteUri + ";tag=" + sipPayloadState.dialog.remoteTag);
                    header.Append("\r\n");
                    //From-header
                    header.Append("From: " + sipPayloadState.dialog.localUri + ";tag=" + sipPayloadState.dialog.localTag);
                    header.Append("\r\n");
                    //AddFromHeader(header, new Uri(sipPayloadState.dialog.localUri));
                    //Call-ID
                    AddCallIDHeader(header, false);
                    //CSeq
                    AddCSeqHeader(SIPRequestType.ACK, header, false);
                    //Content-Length
                    AddContentLength(header, _payloadLength);
                    header.Append("\r\n");
                    // after all said and done when the ACK is dispatched a local current transaction should be destroyed
                    sipPayloadState.transaction = null; // ************************************************************************************** CHECK THIS ******************************
                    break;
                case SIPRequestType.BYE:
                    //Request line
                    AddRequestLine(SIPRequestType.BYE, header, currentToUri);
                    //Via line
                    //AddViaHeader(header, currentFromUri, true);
                    StringBuilder utilitySb = new StringBuilder();
                    utilitySb.Append("Via: SIP/2.0/TCP " + currentFromUri.ToString().Split(':')[1].Split(';')[0] + ";");
                    AddBranchParameter(utilitySb);
                    sipPayloadState.transaction = new SIPTransaction();
                    sipPayloadState.transaction.viaLine = utilitySb.ToString();
                    sipPayloadState.transaction.transactionID = utilitySb.ToString().Split(';')[1].Split('=')[1];
                    sipPayloadState.transaction.method = Enum.GetName(typeof(SIPRequestType), SIPRequestType.BYE);

                    header.Append(utilitySb.ToString());
                    header.Append("\r\n");

                    //Max-Forwards line
                    AddMaxForwards(header);
                    //To line
                    header.Append("To: " + currentToUri.ToString().Split(';')[0] + ";tag=");
                    header.Append(sipPayloadState.dialog.remoteTag);
                    header.Append("\r\n");
                    //From line
                    header.Append("From: " + currentFromUri.ToString().Split(';')[0] + ";tag=");
                    header.Append(sipPayloadState.dialog.localTag);
                    header.Append("\r\n");
                    //Call-ID
                    AddCallIDHeader(header, false);
                    //CSeq line
                    AddCSeqHeader(SIPRequestType.BYE, header, false); // check if the seq number needs to be incremented
                    //Content-Length line                    
                    AddContentLength(header, _payloadLength);
                    header.Append("\r\n");
                    break;                    
            }

            SIPSend(header.ToString(), _payloadByteArray);
        }

        public void SIPSend200Sync() {
            SIPSendResponseSync(SIPResponseType.TWOHUNDRED);
        }

        public void SIPSend400Sync() {
            SIPSendResponseSync(SIPResponseType.FOURHUNDRED);
        }

        public void SIPSend488Sync()
        {
            SIPSendResponseSync(SIPResponseType.FOURHUNDREDANDEIGHTYEIGHT);
        }

        public void SIPSend404Sync()
        {
            SIPSendResponseSync(SIPResponseType.FOURHUNDREDANDFOUR);
        }

        private void SIPSendResponseSync(SIPResponseType code)
        {

            StringBuilder header = new StringBuilder();         

            Log.WriteDebug("UiO.TRAMP.SIP.SIPSendResponseSync(): Entering function for response code '" + Enum.GetName(typeof(SIPResponseType), code) + "' and creating header fields.");

            switch (code)
            {
                case SIPResponseType.TWOHUNDRED:                    
                    //Response line
                    AddResponseLine(SIPResponseType.TWOHUNDRED, header);
                    //via line
                    AddResponseViaLine(header);
                    //to line
                    Uri currentToURI200 = new Uri(sipPayloadState.dialog.localUri);
                    header.Append("To: " + currentToURI200.ToString() + ";");
                    
                    // Very ugly as the entire function is, but for now I'm stuck with it
                    if(sipPayloadState.transaction.method.ToUpper().Equals(Enum.GetName(typeof(SIPRequestType), SIPRequestType.BYE))){
                        header.Append("tag=" + sipPayloadState.dialog.localTag);
                    } else {
                        String rndNumbers200 = GetRandomNumberString(10);
                        header.Append("tag=" + rndNumbers200);
                        sipPayloadState.dialog.localTag = rndNumbers200;                    
                    }
                    header.Append("\r\n");
                    //from line
                    Uri currentFromUri200 = new Uri(sipPayloadState.dialog.remoteUri);
                    header.Append("From: " + currentFromUri200.ToString() + ";");
                    header.Append("tag=" + sipPayloadState.dialog.remoteTag);
                    header.Append("\r\n");
                    //call-id line
                    header.Append("Call-ID: " + sipPayloadState.dialog.callID);
                    header.Append("\r\n");
                    //CSeq line
                    header.Append("CSeq: " + sipPayloadState.dialog.remoteSeqNumber + " " + sipPayloadState.transaction.method);
                    header.Append("\r\n");
                    //contact line
                    if (!(sipPayloadState.transaction.method.ToUpper().Equals(Enum.GetName(typeof(SIPRequestType), SIPRequestType.BYE)))) {
                        header.Append("Contact: <" + sipPayloadState.dialog.localUri.Split(':')[1].Split(';')[0].Trim() + ">");
                        header.Append("\r\n");
                    }
                    
                    //content-type
                    if (!(sipPayloadState.transaction.method.ToUpper().Equals(Enum.GetName(typeof(SIPRequestType), SIPRequestType.BYE)))) {
                        AddContentTypeHeader(header, mediaType);  // for now, this will be variable as SIP will be used as randezvouis protocol for MSRP as well                        
                    }
                    //content-length
                    AddContentLength(header, _payloadLength);
                    header.Append("\r\n");
                    /*
                       Finally, since it is 200OK that hs been returned, according to RFC3261 the current transaction needs to be destroyed, implication of which is 
                     * that ACK that is received shortly after will be a transaction in its own 
                     */
                    sipPayloadState.transaction = null; 
                    break;
                case SIPResponseType.FOURHUNDRED:
                    AddResponseLine(SIPResponseType.FOURHUNDRED, header);
                    //via line
                    AddResponseViaLine(header);
                    //to line
                    Uri currentToURI400 = new Uri(sipPayloadState.dialog.localUri);
                    header.Append("To: " + currentToURI400.ToString() + ";");
                    String rndNumbers400 = GetRandomNumberString(10);
                    header.Append("tag=" + rndNumbers400);
                    sipPayloadState.dialog.localTag = rndNumbers400;                    
                    header.Append("\r\n");
                    //from line
                    Uri currentFromUri400 = new Uri(sipPayloadState.dialog.remoteUri);
                    header.Append("From: " + currentFromUri400.ToString() + ";");
                    header.Append("tag=" + sipPayloadState.dialog.remoteTag);
                    header.Append("\r\n");
                    //call-id line
                    header.Append("Call-ID: " + sipPayloadState.dialog.callID);
                    header.Append("\r\n");
                    //CSeq line
                    header.Append("CSeq: " + sipPayloadState.dialog.remoteSeqNumber + " " + sipPayloadState.transaction.method);
                    header.Append("\r\n");
                    //contact line
                    header.Append("Contact: <" + sipPayloadState.dialog.localUri.Split(':')[1].Split(';')[0].Trim() + ">");
                    header.Append("\r\n");
                    //content-type
                    AddContentTypeHeader(header, mediaType);  // for now, this will be variable as SIP will be used as randezvouis protocol for MSRP as well
                    //content-length
                    AddContentLength(header, _payloadLength);
                    header.Append("\r\n");                    
                    break;
                case SIPResponseType.FOURHUNDREDANDEIGHTYEIGHT:
                    AddResponseLine(SIPResponseType.FOURHUNDREDANDEIGHTYEIGHT, header);
                    //via line
                    AddResponseViaLine(header);
                    //to line
                    Uri currentToURI488 = new Uri(sipPayloadState.dialog.localUri);
                    header.Append("To: " + currentToURI488.ToString() + ";");
                    String rndNumbers488 = GetRandomNumberString(10);
                    header.Append("tag=" + rndNumbers488);
                    sipPayloadState.dialog.localTag = rndNumbers488;                    
                    header.Append("\r\n");
                    //from line
                    Uri currentFromUri488 = new Uri(sipPayloadState.dialog.remoteUri);
                    header.Append("From: " + currentFromUri488.ToString() + ";");
                    header.Append("tag=" + sipPayloadState.dialog.remoteTag);
                    header.Append("\r\n");
                    //call-id line
                    header.Append("Call-ID: " + sipPayloadState.dialog.callID);
                    header.Append("\r\n");
                    //CSeq line
                    header.Append("CSeq: " + sipPayloadState.dialog.remoteSeqNumber + " " + sipPayloadState.transaction.method);
                    header.Append("\r\n");
                    //contact line
                    header.Append("Contact: <" + sipPayloadState.dialog.localUri.Split(':')[1].Split(';')[0].Trim() + ">");
                    header.Append("\r\n");
                    //content-type
                    AddContentTypeHeader(header, mediaType);  // for now, this will be variable as SIP will be used as randezvouis protocol for MSRP as well
                    //content-length
                    AddContentLength(header, _payloadLength);
                    header.Append("\r\n");
                    break;
                case SIPResponseType.FOURHUNDREDANDFOUR:
                    AddResponseLine(SIPResponseType.FOURHUNDREDANDFOUR, header);
                    //via line
                    AddResponseViaLine(header);
                    //to line
                    Uri currentToURI404 = new Uri(sipPayloadState.dialog.localUri);
                    header.Append("To: " + currentToURI404.ToString() + ";");
                    String rndNumbers404 = GetRandomNumberString(10);
                    header.Append("tag=" + rndNumbers404);
                    sipPayloadState.dialog.localTag = rndNumbers404;                    
                    header.Append("\r\n");
                    //from line
                    Uri currentFromUri404 = new Uri(sipPayloadState.dialog.remoteUri);
                    header.Append("From: " + currentFromUri404.ToString() + ";");
                    header.Append("tag=" + sipPayloadState.dialog.remoteTag);
                    header.Append("\r\n");
                    //call-id line
                    header.Append("Call-ID: " + sipPayloadState.dialog.callID);
                    header.Append("\r\n");
                    //CSeq line
                    header.Append("CSeq: " + sipPayloadState.dialog.remoteSeqNumber + " " + sipPayloadState.transaction.method);
                    header.Append("\r\n");
                    //contact line
                    header.Append("Contact: <" + sipPayloadState.dialog.localUri.Split(':')[1].Split(';')[0].Trim() + ">");
                    header.Append("\r\n");
                    //content-type
                    AddContentTypeHeader(header, mediaType);  // for now, this will be variable as SIP will be used as randezvouis protocol for MSRP as well
                    //content-length
                    AddContentLength(header, _payloadLength);
                    header.Append("\r\n");
                    break;
            }

            SIPSend(header.ToString(), _payloadByteArray);
        }

        private void SIPSend(String header, byte[] payload) {
                byte [] headerByteArray = new UTF8Encoding().GetBytes(header);

                if (payload != null) {
                    
                    Log.WriteDebug("UiO.TRAMP.SIP.SIPSend(): Constructing a SIP message with payload to send.");

                    byte[] message = new byte[headerByteArray.Length + payload.Length];
                    Array.Copy(headerByteArray, message, headerByteArray.Length);
                    Array.Copy(payload, 0, message, headerByteArray.Length, payload.Length);

                    Log.WriteDebug("UiO.TRAMP.SIP.SIPSend(): The current SIP message to send has the total of '" + message.Length + "' B.");

                    WriteOutGoingMessage(message);

                    if (tcpConn.SendSync(message)) {
                        Log.WriteLogSuccess("UiO.TRAMP.SIP.SIPSend(): The transport layer sent the current message successfully.");
                    } else {                        
                        Abort();
                    }
                } else {
                    WriteOutGoingMessage(headerByteArray);

                    if (tcpConn.SendSync(headerByteArray)) {
                        Log.WriteLogSuccess("UiO.TRAMP.SIP.SIPSend(): The transport layer sent the current message successfully.");
                    } else {                        
                        Abort();
                    }
                }
        }

        private void AddResponseViaLine(StringBuilder sb)
        {
            sb.Append(sipPayloadState.transaction.viaLine);
            sb.Append("\r\n");
        }

        private void AddContentLength(StringBuilder sb, int length) {
            Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): Adding: " + "Content-Length: " + length.ToString());
            sb.Append("Content-Length: " + length.ToString());
            sb.Append("\r\n");
        }

        private void AddCallIDHeader(StringBuilder sb, Boolean addNew)
        {
            if (addNew)
            {
                String rndString = GetRandomString(17);
                sb.Append("Call-ID: " + rndString);
                sipPayloadState.dialog.callID = rndString;
                Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): Adding: Call-ID: " + rndString);
                sb.Append("\r\n");
            }
            else {
                sb.Append("Call-ID: " + sipPayloadState.dialog.callID);
                Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): Adding: Call-ID: " + sipPayloadState.dialog.callID);
                sb.Append("\r\n");
            }
        }

        private void AddContentTypeHeader(StringBuilder sb, String type){
            Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): Adding: " + "Content-Type: " + type);
            sb.Append("Content-Type: " + type);
            sb.Append("\r\n");
        }

        private void AddContactHeader(StringBuilder sb, Uri currentToUri) {
            Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): Adding: " + "Contact: <" + currentToUri.ToString().Split(';')[0] + ">");
            sb.Append("Contact: <" + currentToUri.ToString().Split(';')[0] + ">");
            sb.Append("\r\n");
        }

        private void AddCSeqHeader(SIPRequestType method, StringBuilder sb, Boolean addNew) {
            if (addNew)
            {
                String rndNumberString = GetRandomNumberString(6);
                sipPayloadState.dialog.localSeqNumber = rndNumberString;
                Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): Adding: " + "CSeq: " + rndNumberString + " " + Enum.GetName(typeof(SIPRequestType), method));
                sb.Append("CSeq: " + rndNumberString + " " + Enum.GetName(typeof(SIPRequestType), method));
                sb.Append("\r\n");
            } else {
                Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): Adding: " + "CSeq: " + sipPayloadState.dialog.localSeqNumber + " " + Enum.GetName(typeof(SIPRequestType), method));
                sb.Append("CSeq: " + sipPayloadState.dialog.localSeqNumber + " " + Enum.GetName(typeof(SIPRequestType), method));
                sb.Append("\r\n");
            }
        }

        private void AddFromHeader(StringBuilder sb, Uri currentFromUri) {
            StringBuilder utilitySb = new StringBuilder();
            utilitySb.Append("From: " + currentFromUri.ToString().Split(';')[0] + ";");
            sipPayloadState.dialog.localUri = currentFromUri.ToString().Split(';')[0];
            AddTagParameter(utilitySb);
            Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): Adding: " + utilitySb.ToString());
            sb.Append(utilitySb.ToString());
            sb.Append("\r\n");
        }

        private void AddTagParameter(StringBuilder sb) {            
            String rndNumbers = GetRandomNumberString(10);
            sipPayloadState.dialog.localTag = rndNumbers;
            sb.Append("tag=" + rndNumbers);
        }
        

        private void AddResponseLine(SIPResponseType type, StringBuilder sb) { 
            switch (type){
                case SIPResponseType.TWOHUNDRED:
                    Log.WriteDebug("UiO.TRAMP.SIP.AddResponseLine(): Adding response line: SIP/2.0 200 OK");
                    sb.Append("SIP/2.0 200 OK");
                    sb.Append("\r\n");
                    break;
                case SIPResponseType.FOURHUNDREDANDFOUR:
                    Log.WriteDebug("UiO.TRAMP.SIP.AddResponseLine(): Adding response line: SIP/2.0 404 NOT FOUND");
                    sb.Append("SIP/2.0 404 NOT FOUND");
                    sb.Append("\r\n");
                    break;
                case SIPResponseType.FOURHUNDREDANDEIGHTYEIGHT:
                    Log.WriteDebug("UiO.TRAMP.SIP.AddResponseLine(): Adding response line: SIP/2.0 488 NOT ACCEPTABLE HERE");
                    sb.Append("SIP/2.0 488 NOT ACCEPTABLE HERE");
                    sb.Append("\r\n");
                    break;
                case SIPResponseType.FOURHUNDRED:
                    Log.WriteDebug("UiO.TRAMP.SIP.AddResponseLine(): Adding response line: SIP/2.0 400 BAD REQUEST");
                    sb.Append("SIP/2.0 400 BAD REQUEST");
                    sb.Append("\r\n");
                    break;
                default:
                    break;
            }
        }       

        private void AddToHeader(StringBuilder sb, Uri currentToUri) {
            Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): Adding: To: '" + currentToUri.ToString().Split(';')[0] + "'");
            sb.Append("To: " + currentToUri.ToString().Split(';')[0]);
            sipPayloadState.dialog.remoteUri = currentToUri.ToString().Split(';')[0];
            sb.Append("\r\n");
        }

        private void AddRequestLine(SIPRequestType method, StringBuilder sb, Uri requestUri) {
            Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): Adding 'Request Line' : " + Enum.GetName(typeof(SIPRequestType), method) + " " + requestUri.ToString() + " " + "SIP/2.0");
            sb.Append(Enum.GetName(typeof(SIPRequestType), method) + " " + requestUri.ToString() + " " + "SIP/2.0");
            if (sipPayloadState.transaction != null)
                sipPayloadState.transaction.method = Enum.GetName(typeof(SIPRequestType), method);
            sb.Append("\r\n");
        }

        private void AddViaHeader(StringBuilder sb, Uri RequestUri, Boolean newTransaction) {
            StringBuilder utilitySb = new StringBuilder();
            
            /*
             *  The way this function is been written right now is messy since the original idea is to reuse it for both INVITE and ACK request.
             *  It was a bad idea! I will leave it this way for now, but it could be subject to change later on 
             */
            
            utilitySb.Append("Via: SIP/2.0/TCP " + RequestUri.ToString().Split(':')[1].Split(';')[0] + ";");

            if (newTransaction) {
                //INVITE, ACK to 200 OK
                AddBranchParameter(utilitySb);
                Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): Adding: " + utilitySb.ToString());

                //INVITE
                if (sipPayloadState.transaction != null) {
                    sipPayloadState.transaction.viaLine = utilitySb.ToString();     // maybe I do not need this one here 
                    sipPayloadState.transaction.transactionID = utilitySb.ToString().Split(';')[1].Split('=')[1];
                }
            } else {
                //ACK to non-200 OK
                utilitySb.Append("branch=" + sipPayloadState.transaction.transactionID);
            } 

            sb.Append(utilitySb.ToString());
            sb.Append("\r\n");
        }

        private void AddBranchParameter(StringBuilder sb) {
            String randomString = GetRandomString(7);
            sb.Append("branch=z9hG4bK" + randomString);

        }

        private void AddMaxForwards(StringBuilder sb) {
            Log.WriteDebug("UiO.TRAMP.SIP.SIPSendRequestSync(): Adding: " + "Max-Forwards: 70");
            sb.Append("Max-Forwards: 70");  //default value, no reason to have anything else in our implementation
            sb.Append("\r\n");
        }        

        private String GetRandomString(int size) {
            Random rnd = new Random((int)DateTime.Now.Ticks);
            StringBuilder sb = new StringBuilder();

            char ch;
            for (int i = 0; i < size; i++)
            {
                ch = Convert.ToChar(Convert.ToInt32(Math.Floor(26 * rnd.NextDouble() + 65)));
                sb.Append(ch);
            }

            return sb.ToString();
        }

        private String GetRandomNumberString(int size)
        {
            string input = "0123456789";
            Random rnd = new Random((int)DateTime.Now.Ticks);

            StringBuilder sb = new StringBuilder();

            char ch;

            for (int i = 0; i < size; i++)
            {
                ch = input[rnd.Next(0, input.Length)];
                sb.Append(ch);
            }
            
            return sb.ToString();
        }

        private void WriteOutGoingMessage(byte[] message)
        {
            StringBuilder utilitySb = new StringBuilder();

            utilitySb.Append(new UTF8Encoding().GetString(message));
            Log.WriteDebug("UiO.TRAMP.SIP.WriteOutGoingMessage(): Sent message: \r\n\r\n" + utilitySb.ToString());
        }
    }
}
