﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Net.Sockets;
using UiO.TRAMP.Logger;
using UiO.TRAMP.SIP;
using System.Net;
using System.IO;
using System.Management;

namespace UiO.TRAMP.Transport
{
    class TCPConnection
    {
        public LoggerClass Log; // this one should be private?
        const int c_number_of_connections_to_accept = 100; // this number was taken from documentation and it really does not mean to much to us

        const int c_send_batch = 8192;
        const int c_receive_buffer_size = 8192;
        const int c_select_read_timeout_micro_seconds = 1000;
        //const int c_maximum_bytes_to_read_in_a_batch = 20480;
        const int c_maximum_bytes_to_read_in_a_batch = 8192;

        public Socket listeningSocket;   
        public Socket currentSocket;

        public TCPConnection (LoggerClass oLog)
        {
            this.Log = oLog;
        }        

        public void ReleaseResources() {
            Log.WriteDebug("UiO.TRAMP.Transport.ReleaseResources(): Releasing resources on transport layer.");
            try {
                if (listeningSocket != null)
                {
                    Log.WriteDebug("UiO.TRAMP.Transport.ReleaseResources(): Shutting down and closing listening socket.");
                    if (listeningSocket.IsBound) {
                        Log.WriteLog("UiO.TRAMP.Transport.ReleaseResources(): Socket is bound.");

                        if (listeningSocket.Connected)
                        {
                            Log.WriteLog("UiO.TRAMP.Transport.ReleaseResources(): Socket is connected. Shutting down and closing.");
                            listeningSocket.Shutdown(SocketShutdown.Both);
                            listeningSocket.Close();
                        }

                        Log.WriteLog("UiO.TRAMP.Transport.ReleaseResources(): Disposing listening socket.");
                        listeningSocket.Dispose();
                    }                    
                }

                if (currentSocket != null)
                {
                    Log.WriteLog("UiO.TRAMP.Transport.ReleaseResources(): Shutting down and closing current socket.");
                    currentSocket.Shutdown(SocketShutdown.Both);
                    currentSocket.Close();

                    Log.WriteLog("UiO.TRAMP.Transport.ReleaseResources(): Disposing current socket.");
                    currentSocket.Dispose();
                }
            } catch (Exception exp) {
                Log.WriteLogWarning("UiO.TRAMP.Transport.ReleaseResources(): An error occured while shutting down and closing sockets. " + exp.ToString());
            }
        }        

        public Boolean ConnectSync(IPAddress ipAddress, int remotePort) {
            IPEndPoint remoteEP;

            try {
                remoteEP = new IPEndPoint(ipAddress, remotePort);
                currentSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);

                currentSocket.NoDelay = true; // don't use Nagle algorithm

                Log.WriteLog("UiO.TRAMP.Transport.ConnectSync(): Connecting to remote IPEndpoint '" + ipAddress + ":" + remotePort + "'.");

                // The thread will block here until the connection is established 
                currentSocket.Connect(remoteEP);

                Log.WriteLogInformational("UiO.TRAMP.Transport.ConnectSync(): Established connection to remote host '" + currentSocket.RemoteEndPoint.ToString() + "'.");

            } catch (Exception exp) {
                Log.WriteLogWarning("UiO.TRAMP.Transport.ConnectSync(): An error occured while connecting to remote host " + exp.ToString());
                return false;
            }

            return true;
        }      

        public Boolean SendSync(byte[] message) { 
            using(MemoryStream ms = new MemoryStream()) {

                int sendBufferSize;
                int bytesReadFromStream;

                int bytesToBeRead;

                byte[] sendBuffer;

                ms.Write(message, 0, message.Length);                
                bytesToBeRead = (int)ms.Length; // the total number of B to send

                Log.WriteDebug("UiO.TRAMP.Transport.SendSync(): Sending total of '" + bytesToBeRead + "' B in current batch.");
                ms.Seek(0, SeekOrigin.Begin);

                while (bytesToBeRead > 0) {                    
                    if (bytesToBeRead < c_send_batch) {
                        sendBufferSize = bytesToBeRead;
                    } else {
                        sendBufferSize = c_send_batch;
                    }

                    Log.WriteDebug("UiO.TRAMP.Transport.SendSync(): Send buffer size is sat to '" + sendBufferSize + "' B.");
                    sendBuffer = new byte[sendBufferSize];

                    bytesReadFromStream = ms.Read(sendBuffer, 0, sendBufferSize);
                    Log.WriteDebug("UiO.TRAMP.Transport.SendSync(): Read '" + bytesReadFromStream + "' B from the current message.");

                    if (bytesReadFromStream == 0) {
                        Log.WriteLogWarning("UiO.TRAMP.Transport.SendSync(): Read 0 bytes from the message stream. Aborting.");
                        return false;
                    }// consider if this situation together with exception should return false which would cause program to stop

                    try {
                        int bytesSentToSocket = currentSocket.Send(sendBuffer, 0, bytesReadFromStream, SocketFlags.None);
                        Log.WriteLogInformational("UiO.TRAMP.Transport.SendSync(): Sent '" + bytesSentToSocket + "' B to the remote host '" + currentSocket.RemoteEndPoint.ToString()  + "'.");

                        if (bytesSentToSocket == 0)
                            throw (new Exception("UiO.TRAMP.Transport.SendAsync(): 0 bytes sent to the socket. Something must be wrong."));

                        if (bytesSentToSocket != sendBufferSize)
                            throw (new Exception("UiO.TRAMP.Transport.SendAsync(): Socket did not accept all the bytes from the send buffer. Since we don't  handle this situation it is a fail condition. Increase the socket buffer size or decrease the sending chunk size."));

                        bytesToBeRead -= bytesReadFromStream;

                    } catch (Exception exp) {
                        Log.WriteLogWarning("UiO.TRAMP.Transport.SendSync(): An error occured while sending data to the socket. " + exp.ToString());
                        return false;
                    }
                }
            }

            return true;
        }

        public Boolean BindAndListenSync (IPAddress ipAddress, int localPort) {
            try
            {
                IPEndPoint localEndPoint = new IPEndPoint(ipAddress, localPort);
                listeningSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);

                Log.WriteLog("UiO.TRAMP.Transport.BindAndListenSync(): Binding to local IPEndpoint '" + ipAddress + ":" + localPort + "'.");

                listeningSocket.Bind(localEndPoint);
                listeningSocket.Listen(c_number_of_connections_to_accept);
                //listeningSocket = listener;
                listeningSocket.NoDelay = true; // don't use Nagle algorithm

                Log.WriteLog("UiO.TRAMP.Transport.BindAndListenSync(): Listening for TCP connections on local IPEndpoint '" + ipAddress + ":" + localPort + "'.");

                currentSocket = listeningSocket.Accept(); // this one blocks

                Log.WriteLogInformational("UiO.TRAMP.Transport.BindAndListenSync(): Received a connection request from a remote host '" + currentSocket.RemoteEndPoint.ToString() + "'.");
            }
            catch (Exception exp) {
                Log.WriteLogWarning("UiO.TRAMP.Transport.BindAndListenSync(): An error occured while waiting for connection from remote host. " + exp.ToString());
                return false;
            }

            return true;
        }

        public Boolean ReceiveSync (MemoryStream rcvStream)
        { 
            byte[] buffer = new byte[c_receive_buffer_size];
            List<Socket> ReadListeners = new List<Socket>();

            int firstTimeCounter = 0;
            int bytesReceived;            

            while (true) {
                try {
                    if (!ReadListeners.Contains(currentSocket))
                        ReadListeners.Add(currentSocket);

                    Socket.Select(ReadListeners, null, null, c_select_read_timeout_micro_seconds);

                    if (ReadListeners.Count > 0) {
                        
                        bytesReceived = ReadListeners[0].Receive(buffer, 0, c_receive_buffer_size, SocketFlags.None);
                        rcvStream.Write(buffer, 0, bytesReceived);
                        
                        Log.WriteLog("UiO.TRAMP.Transport.ReceiveSync(): Received '" + bytesReceived + "' B from remote host '" + ReadListeners[0].RemoteEndPoint.ToString() + "'.");

                        /*
                           We are aborting this read round when there is no more data to read (at the moment) or when the receiving memory stream has reached 'c_maximum_bytes_to_read_in_a_batch' B
                         * The latter condition is used because a potentialy infinite stream could block the entire receive function forever (a possibility for DoS :-) + the parsers need to parse the request and send the answer as soon as possible. 
                         * This also means that protocl parsers will have to reset the position of the utility memory stream so that it always starts from 0                         
                         */

                        if ((ReadListeners[0].Available == 0) || (rcvStream.Length >= c_maximum_bytes_to_read_in_a_batch))
                        {
                            Log.WriteLog("UiO.TRAMP.Transport.ReceiveSync(): Yielding from a current read round (Available = '" + ReadListeners[0].Available + "', Length = '" + rcvStream.Length + "').");    
                            return true;
                        }
                    } else {
                        //timeout timer fired
                        if (firstTimeCounter % 10000 == 0) {
                            //every 10 000th is written to the console
                            Log.WriteLog("UiO.TRAMP.Transport.ReceiveSync(): Waiting for data to receive from remote host.");
                        }

                        firstTimeCounter++;
                    }
                } catch (Exception exp) {
                    Log.WriteLogWarning("UiO.TRAMP.Transport.ReceiveSync(): An error occured while receiving data from remote host. '" + exp.ToString() + "'.");
                    return false;
                }
            }            
        }        
    }    
}
